-- +--------------------------------------------------------------+
-- | Generator of interactive graphical interface for a database  |
-- | Author : Jean-Luc Hainaut                                    |
-- | Created : March 2015                                         |
-- | Last updated : September 2022                                |
-- +--------------------------------------------------------------+

-- delete all user variables to avoid confusion between generator variables
-- and those created by the execution of the DB interface.

showMessage ATTENTION : each table of the database 
            @nMUST have a Primary Key.;
if ('$DIALOGbutton$' <> '$buttonYES$') return;

dropAllVariables;

set debug = 1;

set genVersion = v1.1;

set scriptDir = $scriptDirectory$/BD-2022/Chapitre-19;

set maxSelectWidth = 16000;

-- Select a database --
-- ----------------- --

selectFile db = $dbDirectory$, Database,db;
if ('$db$' = '') return;


-- Compute the name of the prototype --
-- --------------------------------- --

compute localDB = pureNodeName('$db$');
set protoScript = Prototype_$localDB$.sql;
write Generation of prototype $protoScript$;


-- Open the database and its dictionary --
-- ------------------------------------ --

openDB $db$;

createDictionary;

-- ---------------------------------------------------- --
-- ---------------------------------------------------- --
-- Building three metadata temporary tables             --
-- TABLE_TEMP: augmented SYS_TABLE without system table --
-- FKEY_TEMP:  augmented SYS_KEY for FK                 --
-- FIELD_TEMP: fields of the interface                  --
-- ---------------------------------------------------- --
-- ---------------------------------------------------- --

create temporary table TABLE_TEMP(
        TableName      varchar(64),
        TableID        integer, 
        ColList        varchar(2048),
        MaxColLen      integer,
        FieldList      varchar(2048),
        LabelList      varchar(2048),
        PKeyID         integer,
        PKeySize       integer,
        PKeyColList    varchar(256),
        PKeyChar       varchar(512),
        PKeySFW        varchar(512));

--*     TableName      Table name,
--*     TableID        Table ID, 
--*     ColList        List of column names,
--*     MaxColLen      Max length of column names,
--*     FieldList      List of field names,
--*     LabelList      List of labels,
--*     PKeyID         PK ID,
--*     PKeySize       Number of columns of the PK
--*     PKeyColList    PK column list
--*     PKeyChar       Character expression of PK value (cast-cast)
--*     PKeySFW        SFW query returning the list of PK values;

insert into TABLE_TEMP(TableName,TableID,ColList,MaxColLen)
select  T.TableName,T.TableID,
        group_concat(ColName,','),
        max(length(ColName))      
from    SYS_COLUMN C, SYS_TABLE T
where   T.TableID = C.TableID
and     TableType = 'base'
        -- Skip SYS tables
and     T.TableName not like 'SYS_%'
        -- Skip this table
and     T.TableName <> 'TABLE_TEMP'
group by T.TableName,T.TableID
order by T.TableName;

update TABLE_TEMP
set    PKeyID =      (select KeyID 
                      from SYS_KEY 
                      where TableID = TABLE_TEMP.TableID and KeyType = 'PK')

update TABLE_TEMP
set    PKeyColList = (select group_concat(ColName,',') 
                      from   SYS_KEY_COMP
                      where  KeyID = TABLE_TEMP.PKeyID),
       PKeySize =    (select count(*) 
                      from   SYS_KEY_COMP
                      where  KeyID = TABLE_TEMP.PKeyID);

update TABLE_TEMP
set    PKeyChar = 'cast('||replace(PKeyColList,',',' as char)||''-''||cast(')||' as char)';

update TABLE_TEMP
set    PKeySFW = 'select '
               ||case when PKeySize = 1 then PKeyColList else PKeyChar end
               ||' from '||TableName||' order by '||PKeyColList;


-- ======================================================================

create temporary table FKEY_TEMP(
        SourceTable     varchar(64),
        SourceID        integer,
        FKeyID          integer,
        FKeyColList     varchar(2048),
        FKeyName        varchar(2048),
        FKeySize        integer,
        FKeyChar        varchar(512),
        PKeyID          integer,
        TargetTable     varchar(64),
        TargetID        integer);

--*     SourceTable     Name of the source table,
--*     SourceID        ID of the source table,
--*     FKeyID          ID of the FK,
--*     FKeyColList     List of column names of the FK,
--*     FKeyName        FK name = concat. of column names,
--*     FKeySize        Number of columns of the FK,
--*     FKeyChar        Character expression of FK value (cast-cast),
--      PKeyID          ID of the target primary key, (not used yet)
--*     TargetTable     Name of the target table,
--*     TargetID        ID of the target table

insert into FKEY_TEMP(SourceTable,SourceID,FKeyID,PKeyID,TargetTable,TargetID)
select SourceTable,ST.TableID,SourceKey,TargetKey,TargetTable,TT.TableID
from   SYS_INTER_TABLE IT,TABLE_TEMP ST,SYS_TABLE TT
where  IT.SourceTable = ST.TableName
and    IT.TargetTable = TT.TableName;

update FKEY_TEMP
set    FKeyColList = (select group_concat(ColName,',') 
                      from   SYS_KEY_COMP
                      where  KeyID = FKEY_TEMP.FKeyID),
       FKeyName =    (select group_concat(ColName,'-') 
                      from   SYS_KEY_COMP
                      where  KeyID = FKEY_TEMP.FKeyID),
       FKeySize =    (select count(*) 
                      from   SYS_KEY_COMP
                      where  KeyID = FKEY_TEMP.FKeyID);

update FKEY_TEMP
set    FKeyChar = 'cast('||replace(FKeyColList,',',' as char)||''-''||cast(')||' as char)';


-- ======================================================================

create temporary table FIELD_TEMP(
        TableID        integer, 
        FieldName      varchar(64),
        Nullable       char(1),
        FieldOrder     integer,
        TargetKey      integer);

--*     TableID        Table ID, 
--*     FieldName      Field name,
--*     Nullable       Is the field nullable or not,
--*     FieldOrder     Sequence oder of the field (from column order),
--*     TargetKey      PKey ID of the target table,

insert into FIELD_TEMP(TableID,FieldName,Nullable,FieldOrder,TargetKey)
   -- First insert concatenated FK columns --
select T.TableID,
       group_concat(C.ColName,'-'),
       min(SK.KeyNull),
       min(C.ColSeq) as FieldOrder,
       min(TKC.KeyID)
from   SYS_COLUMN C, SYS_KEY_COMP SKC, TABLE_TEMP T, SYS_KEY SK, SYS_KEY_COMP TKC 
where  C.ColID   = SKC.ColID
and    C.TableID = T.TableID
and    SKC.KeyID  = SK.KeyID
and    SK.KeyType = 'FK'
and    TKC.KeyCompID = SKC.CompTarget
group by T.TableName,T.TableID,SKC.KeyID
   union
   -- Now insert all the other columns --
select T.TableID,C.ColName,'X',C.ColSeq as FieldOrder,null
from   SYS_COLUMN C,TABLE_TEMP T
where  T.TableID = C.TableID
and    C.ColID not in (select ColID from SYS_KEY_COMP where CompTarget is not null);

-- mark nullable FK --
update FIELD_TEMP
set    Nullable = (select ColNull from SYS_COLUMN 
                   where  ColName = FIELD_TEMP.FieldName
                   and    TableID = FIELD_TEMP.TableID)
where Nullable = 'X';


-- Completing TABLE_TEMP(FieldList,LabelList) --

update TABLE_TEMP
set    FieldList = (select group_concat(FT.FieldName,',') 
                    from   (select FieldName from FIELD_TEMP
                            where  TableID = TABLE_TEMP.TableID
                            order by FieldOrder) FT
                    );

update TABLE_TEMP
set    LabelList = (select group_concat(case when Nullable = 'N'
                                             then FT.FieldName||'*:'
                                             else FT.FieldName||'  :'
                                        end
                                      ||case when FT.TargetKey is not null
                                             then '['||(select TT.PKeySFW
                                                        from   TABLE_TEMP TT
                                                        where  TT.TableID = FT.TargetKey)
                                                     ||']'
                                             else ''
                                        end,'|') 
                    from   (select FieldName,Nullable,TargetKey from FIELD_TEMP
                            where  TableID = TABLE_TEMP.TableID
                            order by FieldOrder) FT
                    );


-- ---------------------------- --
-- Metadata created             --
-- ---------------------------- --


-- Dump of the three metadata tables --
-- --------------------------------- --
if ($debug$ = 1);
  write-b TABLE_TEMP;
  select TableName,TableID,ColList,MaxColLen,FieldList from TABLE_TEMP;
  select TableName,TableID,LabelList from TABLE_TEMP;
  select TableName,TableID,PKeyID,PKeySize,PKeyColList,PKeyChar,PKeySFW from TABLE_TEMP;

  write-b FKEY_TEMP;
  select SourceTable,SourceID,FKeyID,FKeyColList,FKeyName,FKeySize,FKeyChar,
          PKeyID,TargetTable,TargetID
  from FKEY_TEMP;

  write-b FIELD_TEMP;
  select * from FIELD_TEMP;
  write;
endif;


-- --------------------------------------------- --
-- --------------------------------------------- --
-- Generating the prototype DB manager           --
-- --------------------------------------------- --
-- --------------------------------------------- --

-- Reroot writing to the interface scrit file  --
-- ------------------------------------------- --

outputOpen $scriptDir$/$protoScript$;

write;
write -- -----------------------------------------------------------------------------;
write -- Prototype manager of database $localDB$.;
write -- Date : $date$ - $time$;
write -- Generator version $genVersion$;
write -- -----------------------------------------------------------------------------;
write;

-- ---------------------- --
-- PROLOGUE GENERATION    --
-- ---------------------- --

write;
write openDB $db$;;
write;

write onError continue;;
write;

write -- Display parameters --;
write set defaultLimit = 50;;
write set defaultxText = 72;;
write set defaultyText = 24;;
write;

write set limit = $defaultLimit$;;
write set xText = $defaultxText$;;
write set yText = $defaultyText$;;
write;

write -- Running parameters --;
write set append = 0;;
write set exploreColumns = 1;;
write set execMode = 1;;
write set result = ;;
write;

write -- Register of table PK --;
for TableName = [select TableName from TABLE_TEMP order by TableName]
    write set id$TableName$ = ;;

write;
write -- --------------------------------- --;
write -- Command panel and rooting         --;
write -- --------------------------------- --;
write;
write set oper = 1;;
write while (True);;
write @S3label SELECT;;

write @tparameter styleBoxFrame = relief;;
write @tparameter styleBoxMessage = flat;;


-- ----------------------------------- --
-- GENERATION of OPERATIONS  BOX       --
-- ----------------------------------- --

-- Build operation columns --

extract list1 = select group_concat(TableName,'|')
                from (select * from TABLE_TEMP order by TableName);

-- Generate Main box --

write @taskCombo [selectOne-u oper = {Create rows}  |$list1$@n
                             @t@t||{Modify rows} |$list1$@n
                             @t@t||{Delete rows}|$list1$@n
                             @t@t||{List tables}   |$list1$@n
                             @t@t||{Explore tables} |$list1$]@n
            @t@t|[ask-u limit,xText,yText = [/bParameters of function "List tables"]@n
                                       @t@t Max number of data rows to extract (1-100000)@n
                                       @t@t|Window width (24-132)@n
                                       @t@t|Window height (6-72)]@n
            @t@t|[selectOne-u exploreColumns = [/bParameters of function "Explore tables".
                                                  @S3Data are presented:]@n
                                       @t@t in 1 column||in 3 columns]@n
            @t@t|[selectOne-u execMode = [/bExecution of updates]@n
                                       @t@t test mode||real mode]@n
            @t@t|[selectMany-u append = Append results];;

-- Generate validation --

substitution off;
write @tif ('$DIALOGbutton$' = 'Cancel') exit;;
write @t@tif (isNotNumeric('$limit$'))  set limit = $defaultLimit$;;
write @t@tif (isNotNumeric('$xText$'))  set xText = $defaultxText$;;
write @t@tif (isNotNumeric('$yText$'))  set yText = $defaultyText$;;
write @t@tif ($limit$ not between  1  and 100000) set limit = $defaultLimit$;;
write @t@tif ($xText$ not between  24 and 132)    set xText = $defaultxText$;;
write @t@tif ($yText$ not between  6  and 72)     set yText = $defaultyText$;;
write @t@tif ($exploreColumns$ = 1) set -sep = |;;
write @t@tif ($exploreColumns$ = 2) set -sep = ||;;
write @t@tif ($append$ = 0) set outMode = outputOpen;;
write @t@tif ($append$ = 1) set outMode = outputAppend;;
substitution on;

-- --------------------------- --
-- GENERATION OF ROOTING       --
-- --------------------------- --

set num = 0;

-- CREATION routing --

for name = [select TableName from TABLE_TEMP order by TableName];
   compute num = $num$ + 1;
   write @tif ($oper$ = $num$) goto create-$name$;;
endfor;

-- Modification rooting --

for name = [select TableName from TABLE_TEMP order by TableName];
   compute num = $num$ + 1;
   write @tif ($oper$ = $num$) goto modify-$name$;;
endfor;

-- DELETE routing --

for name = [select TableName from TABLE_TEMP order by TableName];
   compute num = $num$ + 1;
   write @tif ($oper$ = $num$) goto delete-$name$;;
endfor;

-- LIST routing --

for name = [select TableName from TABLE_TEMP order by TableName];
   compute num = $num$ + 1;
   write @tif ($oper$ = $num$) goto list-$name$;;
endfor;

-- EXPLORE routing --

for name = [select TableName from TABLE_TEMP order by TableName];
   compute num = $num$ + 1;
   write @tif ($oper$ = $num$) goto explore-$name$;;
endfor;

-- Routing for EXPLORE continuation --

write @S3label VISIT;;

for name = [select TableName from TABLE_TEMP order by TableName]
   write @tif ('$id$name$$' <> '') goto explore-$name$;;

write endwhile;;

-- ---------------------------- --
-- CLOSING MAIN PROCEDURE       --
-- ---------------------------- --

write onError stop;;
-- 'abortDB' for testing. To remplace by 'commitDB' in operational version.
write abortDB;;
write closeDB;;
write stop;;
write;



write -- ------------------------------------- --;
write -- GENERATION OF CREATION OPERATIONS     --;
write -- ------------------------------------- --;
write;
for TableID,TableName,FieldList,LabelList
    = [select TableID,TableName,FieldList,LabelList
       from TABLE_TEMP order by TableName];

   write -------------------------;
   write label create-$TableName$;;
   write -------------------------;
   write set operation = CREATE;;
   write set currentTable = $TableName$;;
   write;

   write while (True);;


  -- --------------------------------- --
  -- Build the data entry dialogue box --
  -- --------------------------------- --

   -- Generate the 'ask' data entry box --
   write @S3ask $FieldList$ = [/bCreate $TableName$] $LabelList$;;

   -- if ('$DIALOGbutton$' = 'Cancel') exit; --
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel') exit;;
   substitution on;


  -- --------------------------------- --
  -- Build the data insert statement   --
  -- --------------------------------- --

  -- Generate extraction of columns of composite FKs --
  -- ----------------------------------------------- --

   for  FKeyColList,PKeyColList,TargTableName,PKeyChar,FKeyName
                                  = [select FK.FKeyColList,T.PKeyColList,T.TableName,T.PKeyChar,FK.FKeyName
                                     from   FKEY_TEMP FK, TABLE_TEMP T
                                     where  FK.TargetID = T.TableID
                                     and    FK.FKeySize > 1
                                     and    FK.SourceID = $TableID$];

      --       extract NCOM,NPRO     = select NCOM,NPRO     from DETAIL          where ...||...   = '$NCOM-NPRO$';
      write @S3extract $FKeyColList$ = select $PKeyColList$ from $TargTableName$ where $PKeyChar$ = '$$FKeyName$$';;

   endfor;


  -- Generate value list of "insert" query --
  -- ------------------------------------- --
  -- Rules: - if nullable,     numeric,       empty: set to 'null'
  --        - if nullable,     character,     empty: set to 'null'
  --        - if nullable,     numeric,   not-empty: as is
  --        - if nullable,     character, not-empty: as is, quoted
  --        - if not-nullable, numeric,       empty: set to 'null' --> will trigger an SQL error
  --        - if not-nullable, character,     empty: set to 'null' --> will trigger an SQL error
  --        - if not-nullable, numeric,   not-empty: as is
  --        - if not-nullable, character, not-empty: as is, quoted
  -- (Note that a not-nullable character column is considered as = null. Can be changed)


  -- Generate double-quoting character values                       --
  -- -------------------------------------------------------------- --
  -- Superseded by '' substitution operators
  -- for CColName = [select ColName from SYS_COLUMN
  --                 where TableID = $TableID$
  --                 and   ColType in ('char','varchar','date','text','time','datetime','clob')]
  --    write @S3function $CColName$ = LStr:SQLquote2 {$$CColName$$};;

  -- Set numeric empty column values to null (even if not-nullable) --
  -- -------------------------------------------------------------- --
   for ColName = [select ColName from SYS_COLUMN
                  where  TableID = $TableID$
                  and    ColType not in ('char','varchar','date','text','time','datetime','clob')]
       write @S3if ('$$ColName$$' = '') set $ColName$ = null;;

  -- Build the value list of the insert query                       --
  -- -------------------------------------------------------------- --
   extract values =
          select group_concat(case 
                              when ColType in ('char','varchar','date','text','time','datetime','clob')
                              --then 'case when length(''$'||ColName||'$'') = 0 then null else '
                              then 'case when '''||ColName||''' = '''' then null else '
                                          ||case
                                            when ColType in ('char','varchar','date','text','time','datetime','clob')
                                            then ''''||ColName||''''
                                            else '$'||ColName||'$'
                                            end
                                          ||' end'
                              else '$'||ColName||'$'
                              end,
                              ',')
          from (select * from SYS_COLUMN
                where  TableID = $TableID$
                order by ColSeq);

-- In this alternate version, empty strings are valid values for character not-nullable columns --
-- -------------------------------------------------------------------------------------------- --
--   extract values =
--          select group_concat(case 
--                              when ColNull = 'Y'
--                             -- generate value of a nullable column
--                              then 'case when length(''$'||ColName||'$'') = 0 then null else '
--                                          ||case
--                                            when ColType in ('char','varchar','date','text','time','datetime','clob')
--                                            then '''$'||ColName||'$'''
--                                            else '$'||ColName||'$'
--                                            end
--                                          ||' end'
--                             -- generate value of a not-nullable column
--                              else case
--                                   when ColType in ('char','varchar','date','text','time','datetime','clob')
--                                   then '''$'||ColName||'$''' else '$'||ColName||'$'
--                                   end
--                              end,
--                              ',') as list3
--          from (select * from SYS_COLUMN
--                where TableID = $tid$
--                order by ColSeq);

  -- Generate the "insert" query --
  -- --------------------------- --
   write @S3insert into $TableName$ values($values$);;

  -- Generate error management --
  -- ------------------------- --
  substitution off;
   write @S3if ('$SQLdiag$' <> 'OK');;
   substitution on;
   write @S6execSQL $scriptDir$/ProcessErrorSQL.sql;;
   write @S3else;;
   write @S6if ($execMode$ = 2) commitDB;;
   write @S3endif;;

  -- Generate End of insert loop --
  -- --------------------------- --
   write endwhile;;

  -- Generate Return to main box --
  -- --------------------------- --
   write goto SELECT;;
   write;

endfor;
write;


write -- ----------------------------------------- --;
write -- GENERATION OF MODIFICATION OPERATIONS     --;
write -- ----------------------------------------- --;
write;
for TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
    = [select TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
       from TABLE_TEMP order by TableName];

   write -----------------------;
   write label modify-$TableName$;;
   write -----------------------;
   write set operation = MODIFY;;
   write set currentTable = $TableName$;;
   write;

   write while (True);;

  -- --------------------------------- --
  -- Build row selection dialogue box --
  -- --------------------------------- --

   -- Generate the 'ask' row selection box --
   write @S3ask id = [/bModify $TableName$] $TableName$:[$PKeySFW$];;
   write;

   -- if ('$DIALOGbutton$' = 'Cancel') exit; --
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel') exit;;
   substitution on;
   write;

   -- Generate doublequoting if needed --
   -- Superseded by '' substitution operators
   -- write @S3function id = LStr:SQLquote2 {$id$};;

   -- Extract data from selected row --
   write @S3extract $ColList$ = select $ColList$ from $TableName$ where $PKeyChar$ = 'id';;
   write;

   -- Generate composite key building
   for FKeyName,FKeyConcat,KeyNull
                   = [select FK.FKeyName,
                             ''||replace(FK.FKeyName,'-','-')||'',
                             K.KeyNull
                      from   FKEY_TEMP FK,SYS_KEY K
                      where  FK.SourceID = $TableID$
                      and    FK.FKeySize > 1
                      and    K.KeyID = FK.FKeyID];
      write @S3set $FKeyName$ = FKeyConcat;;
      if ('$KeyNull$' = 'Y');
         -- Dirty/cheap trick to check whether a FK value includes null values
         -- The value cannot include '^^^^'.
         write @S3compute testNull = '^^'||replace('FKeyName','-','^^-^^')||'^^';;
         write @S3if ('$testNull$' like '%^^%');
         write @S6set $FKeyName$ = ;;
      endif;
   endfor;

   -- Generate the 'ask-u' data modification box --
   write-b @S3ask-u $FieldList$ = [/bModify $TableName$] $LabelList$;;

   -- if ('$DIALOGbutton$' = 'Cancel') exit; --
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel') exit;;
   substitution on;
   write;

   -- Generate explosion of components of composite FK --
   for FKeyColList,TPKeyColList,TTableName,TPKeyChar,FKeyConcat
                           = [select FK.FKeyColList,TT.PKeyColList,TT.TableName,TT.PKeyChar,
                              '$'||replace(FK.FKeyName,'-','$-$')||'$'
                              from  FKEY_TEMP FK,TABLE_TEMP TT
                              where FK.SourceID = $TableID$
                              and   TT.TableID = FK.TargetID
                              and   FK.FKeySize > 1]
      write @S3extract $FKeyColList$ = select $TPKeyColList$ from $TTableName$
                                       where $TPKeyChar$ = '$FKeyConcat$';;

   -- Generate double-quoting character values --
   -- Superseded by '' substitution operators
   --   for CColName = [select ColName from SYS_COLUMN
   --                where TableID = $TableID$
   --                and   ColType in ('char','varchar','date','text','time','datetime','clob')]
   --   write @S3function $CColName$ = LStr:SQLquote2 {$$CColName$$};;

   -- Generate null for numeric empty values --
   for CColName = [select ColName from SYS_COLUMN
                   where  TableID = $TableID$
                   and    ColType not in ('char','varchar','date','text','time','datetime','clob')]
      write @S3if ('$CColName$' = '') set $CColName$ = null;;
   write;

   -- Generate data update statement --
   extract setList = select group_concat(
                     '@n@S10'||ColName||' = '||case when ColType in ('char','varchar','date','text','time','datetime','clob')
                          then 'case when '''||ColName||''' = '''' then null else '''||ColName||''' end'
                          else ''||ColName||''
                     end,',')
                     from SYS_COLUMN
                     where TableID = $TableID$;
   write @S3update $TableName$ set $setList$;
   write @S3where $PKeyChar$ = 'id';;
   write;

  -- Generate error management --
  -- ------------------------- --
  substitution off;
   write @S3if ('$SQLdiag$' <> 'OK');;
   substitution on;
   write @S6execSQL $scriptDir$/ProcessErrorSQL.sql;;
   write @S3else;;
   write @S6if ($execMode$ = 2) commitDB;;
   write @S3endif;;

  -- Generate End of insert loop --
  -- --------------------------- --
   write endwhile;;
   write;

  -- Generate Return to main box --
  -- --------------------------- --
   write goto SELECT;;
   write;

endfor;
write;


write -- ----------------------------------------- --;
write -- GENERATION OF DELETION OPERATIONS         --;
write -- ----------------------------------------- --;
write;
for TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
    = [select TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
       from   TABLE_TEMP order by TableName];

   write -----------------------;
   write label delete-$TableName$;;
   write -----------------------;
   write set operation = DELETE;;
   write set currentTable = $TableName$;;
   write;

   write while (True);;

  -- --------------------------------- --
  -- Build row selection dialogue box --
  -- --------------------------------- --

   -- Generate the 'ask' row selection box --
   write @S3ask id = [/bDelete $TableName$] $TableName$:[$PKeySFW$];;

   -- if ('$DIALOGbutton$' = 'Cancel') exit; --
   -- -------------------------------------- --
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel') exit;;
   substitution on;
   write;

   -- Generate doublequoting if needed --
   -- -------------------------------- --
    -- Superseded by '' substitution operators
   --write-b -- Doublequoting "id" (just in case!) --;
   --write @S3function id = LStr:SQLquote2 {$id$};;
   --write;

   -- Extract data from selected row --
   write @S3extract $ColList$ = select $ColList$ from $TableName$ where $PKeyChar$ = 'id';;
   write;

   -- Generate confirmation dialogue --
      -- Transform column name list into label list --
   compute pureLabels = replace('$ColList$', ',', ':|')||':';
      -- Generate confirmation box --
   write @S3showData $ColList$ = [/bConfirm or cancel deletion] $pureLabels$;;
      -- Exit if cancel --
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel') exit;;
   substitution on;

   -- Generate delete statement --
   -- ------------------------- --
   write @S3delete from $TableName$ where $PKeyChar$ = 'id';;

  -- Generate error management --
  -- ------------------------- --
   substitution off;
   write @S3if ('$SQLdiag$' <> 'OK');;
   substitution on;
   write @S6execSQL $scriptDir$/ProcessErrorSQL.sql;;
   write @S3else;;
   write @S6if ($execMode$ = 2) commitDB;;
   write @S3endif;;

  -- Generate End of insert loop --
  -- --------------------------- --
   write endwhile;;
   write;

  -- Generate Return to main box --
  -- --------------------------- --
   write goto SELECT;;
   write;

endfor;
write;


write -- ------------------------------------ --;
write -- GENERATION OF LIST OPERATIONS        --;
write -- ------------------------------------ --;
write;
for TableName= [select TableName from TABLE_TEMP order by TableName];
   write -----------------------;
   write label list-$TableName$;;
   write -----------------------;
   write set operation = LIST;;
   write set currentTable = $TableName$;;
   write;

   substitution off;
   write $outMode$ result.var;;
   substitution on;
   write;

   write write-b Table $TableName$;;
   write select * from $TableName$ limit $limit$;;
   write showText result = [/w0/x$xText$/y$yText$/bContents of table $TableName$];;
   write outputAppend console;;
   write;
   write goto SELECT;;
   write;
endfor;
write;


write -- --------------------------------------- --;
write -- GENERATION OF EXPLORE OPERATIONS        --;
write -- --------------------------------------- --;
write;
for TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
    = [select TableID,TableName,ColList,FieldList,LabelList,PKeyChar,PKeySFW
       from TABLE_TEMP order by TableName];

   write --------------------------;
   write label explore-$TableName$;;
   write --------------------------;
   write set operation = EXPLORE;;
   write set currentTable = $TableName$;;
   write;

   write parameter styleGeneralMessage = relief;;
   write if ($exploreColumns$ = 1) parameter styleBoxFrame = relief;;
   write if ($exploreColumns$ = 2) parameter styleBoxFrame = flat;;
   write if ($exploreColumns$ = 1) parameter styleBoxMessage = flat;;
   write if ($exploreColumns$ = 2) parameter styleBoxMessage = relief;;
   write parameter styleEntry = flat;;
   write parameter styleText = flat;;

   write-b set id = $id$TableName$$;;

   write-b -- New or undergoing exploration? --;
   write if ('$id$' = '');; 
   write @S3ask id = [/bExplore $TableName$] $TableName$:[$PKeySFW$];
   substitution off;
   write @S3if ('$DIALOGbutton$' = 'Cancel' or '$id$' = '') goto SELECT;;
   substitution on;
   write endif;;

   -- write-b -- Doublequoting "id" (just in case!) --;
   -- Superseded by '' substitution operators
   -- write function id = LStr:SQLquote2 {$id$};;
   
   -- Initialize askCombo box definition and decoding statements --
   set box = askCombo [/bEnvironnement of $TableName$ $id$]@n@S9;
   set decode = ;

   write-b -- Extract data from selected $TableName$ row --;
   write extract $ColList$  @n@S8= select $ColList$ 
                            @n@S10from $TableName$ 
                            @n@S10where $PKeyChar$ = '$id$';;

   -- Add to askCombo box the display of current row --
   compute pureLabels = replace('$ColList$', ',', ':|')||':';
   set box = $box$[/y10showData $ColList$ = [/bCurrent $TableName$ row] $pureLabels$];

   -- Compute maximum length of all column names --
   extract MaxLen = select max(MaxColLen) from TABLE_TEMP
                    where TableID in (select TargetID from FKEY_TEMP where SourceID = $TableID$);

   write-b -- Extract data from referenced rows   --;
   set firstBox = 1;
   for -FKeyID,-FKeyColList,-ColList,-TableName,-PKeyChar,-PKeyColList,-MaxColLen
               = [select FK.FKeyID,FK.FKeyColList,T.ColList,T.TableName,T.PKeyChar,T.PKeyColList,T.MaxColLen 
                  from   FKEY_TEMP FK,TABLE_TEMP T
                  where  FK.SourceID = $TableID$
                  and    FK.TargetID = T.TableID];

      compute -ColList1 = substr(replace('$-ColList$,',',','-$-FKeyID$,'),-1,-10000);
      compute -FKeyColList1 = '$'||replace('$-FKeyColList$', ',', '$-$')||'$';

      write extract $-ColList1$ @n@S8= select $-ColList$ 
                                @n@S10from $-TableName$
                                @n@S10where $-PKeyChar$ = '$-FKeyColList1$';;

      -- Define the label offset --
      compute offset = int(($MaxLen$ - $-MaxColLen$)*1.8) + 1;
      set off = @@S$offset$;

      -- Add to askCombo box the display of targeted row --
      compute pureLabels = '$off$'||replace('$-ColList$', ',', ':|$off$')||':';
      if ($firstBox$ = 1);
         -- Skip to next column if first box --
         set box = $box$@n@S9$-sep$[showData $-ColList1$
                           = [/b$-TableName$ referenced by $TableName$ via ($-FKeyColList$)] $pureLabels$];
         set firstBox = 0;
      else;
         set box = $box$@n@S9|[showData $-ColList1$
                           = [/b$-TableName$ referenced by $TableName$ via ($-FKeyColList$)] $pureLabels$];
      endif;

      -- Add to askCombo box the navigation button --
      set box = $box$@n@S9|[selectMany check$-TableName$ = Visit $-TableName$];

      -- Add to decode list the button conversion statements --
      compute PK = '$'||replace('$-PKeyColList$', ',', '-$-FKeyID$$-$')||'-$-FKeyID$$';
      set decode = $decode$@nif ($check$-TableName$$ = 1);;
      set decode = $decode$@n@S3set id$-TableName$ = $PK$;;
      set decode = $decode$@nendif;;
   endfor;

   write-b -- Extract data from referencing rows --;
   set firstBox = 1;
   for -TableName,-FKeyID,-FKeyChar,-FKeyColList,-PKeyChar
               = [select T.TableName,FK.FKeyID,FK.FKeyChar,FK.FKeyColList,T.PKeyChar 
                  from   FKEY_TEMP FK,TABLE_TEMP T
                  where  FK.TargetID = $TableID$
                  and    FK.SourceID = T.TableID];

      -- Store extracted data in a text variable --
      write outputOpen -$-TableName$-$-FKeyID$.var;;
      write select * from $-TableName$ where $-FKeyChar$ = '$id$';;
      write outputAppend console;;

      -- Add to askCombo box the display of dependent rows --
      if ($firstBox$ = 1);
         -- Skip to next column if first box --
         set box = $box$@n@S9$-sep$[showText -$-TableName$-$-FKeyID$
                                  = [/b/w0/x48/y6$-TableName$ referencing $TableName$ via ($-FKeyColList$)]];
         set firstBox = 0;
      else;
         set box = $box$@n@S9|[showText -$-TableName$-$-FKeyID$
                           = [/b/w0/x48/y6$-TableName$ referencing $TableName$ via ($-FKeyColList$)]];
      endif;

      -- Add to askCombo box the navigation listbox --
      compute -FKeyColList = '$'||replace('$-FKeyColList$', ',', '$-$')||'$';
      set box = $box$@n@S9|[ask id$-TableName$ = Visit $-TableName$:[select $-PKeyChar$ 
                                                                     from $-TableName$
                                                                     where $-FKeyChar$ = 'id']];
   endfor;

   write-b -- Empty navigation registers --;
   for TName = [select distinct TName
                from (select TargetTable as TName from FKEY_TEMP where SourceID = $TableID$
                      union
                      select SourceTable as TName from FKEY_TEMP where TargetID = $TableID$
                      union
                      select '$TableName$')
                order by TName]
      write set id$TName$ = ;;

   write-b -- Building the display box --;
   write $box$;;
   if (length('$decode$') > 0)
       write $decode$;

   -- if ('$DIALOGbutton$' = 'Cancel') goto SELECT) --
   substitution off;
   write if ('$DIALOGbutton$' = 'Cancel') goto SELECT;;
   substitution on;

   write-b -- Continue the exploration --;
   write-a goto VISIT;;

endfor;

outputAppend window;

--select * from FIELD_TEMP;
--select TableName,TableID,ColumnList,PrefColumnList,PKeyID,PKcolNbr from TABLE_TEMP;
--select TableName,TableID,PKeyComp,CharPKey,PKeySFW from TABLE_TEMP;

closeDB;

write Generation completed;
