createOrReplaceDB RETAIL.db;

create table CUSTOMER (
        CustID  char(10) not null,
        Name    char(32) not null,
        Address char(60) not null,
        City    char(30) not null,
        Account decimal(9,2) not null default 0.0,
        constraint PKCUST primary key (CustID));

create table ITEM (
        ItemID      char(15) not null,
        Description char(60) not null,
        Price       decimal(4,2) not null,
        SuppID      char(10),
        SuppPrice   decimal(4,2),
        QonHand     integer not null,
        Qord        integer default 0,
        Qavail      integer not null generated always as (QonHand+Qord),
        Qeco        integer not null,
        constraint PKITEM primary key (ItemID));

create table CUSTORDER (
        OrdID   integer  not null,
        DateOrd date     not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2),
        Qty     integer  not null,
        State   char(20) not null default 'recorded',
        constraint PKCORD primary key (OrdID),
        constraint FKORDCUST foreign key (CustID)
           references CUSTOMER
           on delete no action on update cascade,
        constraint FKORDITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade,
        constraint CHKQTY check(Qty > 0));

create table CUSTINVOICE (
        InvID   integer  not null,
        DateInv date     not null,
        OrdID   integer  not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Qty     integer  not null,
        Amount  integer  not null generated always as (Qty*Price),
        State   char(20) not null default 'recorded',
        constraint PKCINV primary key (InvID),
        constraint FKINVORD foreign key (OrdID)
           references CUSTORDER
           on delete no action on update cascade);

create table SUPPLIER (
        SuppID  char(10) not null,
        Name    char(32) not null,
        City    char(30) not null,
        constraint PKSUPP primary key (SuppID));

create table OFFER (
        SuppID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Delay   integer not null,
        constraint PKOFF primary key (SuppID,ItemID),
        constraint FKOFFSUPP foreign key (SuppID)
           references SUPPLIER
           on delete no action on update cascade,
        constraint FKOFFITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade);

create table SUPPORDER (
        OrdID   integer not null,
        DateOrd date not null,
        ItemID  char(15) not null,
        SuppID  char(10),
        Price   decimal(4,2),
        Qty     integer  not null,
        State  varchar(20) not null default 'recorded',
        constraint PKSORD primary key (OrdID)
        constraint FKORDSUPP foreign key (SuppID,ItemID)
           references OFFER
           on delete no action on update cascade);

-- Variables used in CUSTORDER processing
create table V_CORDER as
    select Price,Qavail,0 as mQeco,0 as InvID
    from   ITEM where False;
create table TLEVEL as select 0 as Level;;

set ; = ;;
set message   = select printMessage;
set writeFile = select raise(IGNORE) where writeFile;


-- CUSTOMER
-- create trigger TRG_CUSTOMER_UPD     (empty)

-- CUSTORDER
-- create trigger TRG_CUSTORDER_INS1
-- create trigger TRG_CUSTORDER_INS2
-- create trigger TRG_CUSTORDER_UPD    (empty)

-- ITEM
-- create trigger TRG_ITEM_UPD1        (empty)
-- create trigger TRG_ITEM_UPD2

-- CUSTINVOICE
-- create trigger TRG_CUSTINVOICE_INS  (empty)
-- create trigger TRG_CUSTINVOICE_UPD1 (empty)
-- create trigger TRG_CUSTINVOICE_UPD2
-- create trigger TRG_CUSTINVOICE_UPD3

-- SUPPORDER
-- create trigger TRG_SUPPORDER_INS    (empty)
-- create trigger TRG_SUPPORDER_UPD


-- ============================================ --
-- TRIGGERS on CUSTOMER                         --
-- ============================================ --

create trigger TRG_CUSTOMER_UPD
after update on CUSTOMER
for each row
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update CUSTOMER set '
           ||rtrim(iif(new.CustID = old.CustID, '', 'CustdID = '||''''||new.CustID ||''''||',')
                 ||iif(new.Name = old.Name, '', 'Name = '       ||''''||new.Name   ||''''||',')
                 ||iif(new.Address = old.Address,'','Address = '||''''||new.Address||''''||',')
                 ||iif(new.City = old.City, '', 'City = '         ||''''||new.City   ||''''||',')
                 ||iif(new.Account = old.Account, '', 'Account = '||new.Account), ',') ) $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTOMER_UPD') $;$
   -- -----------------------------------------------------
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTOMER_UPD') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- ====================================================== --
-- TRIGGERS on CUSTORDER                                  --
-- ====================================================== --

-- ====================================================== --
-- INSERT: "Quantity on hand" is sufficient               --
-- ====================================================== --

create trigger TRG_CUSTORDER_INS1
after insert on CUSTORDER
for each row
when new.Qty <= (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Extract initial values used more than once in this trigger
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: insert into CUSTORDER ('
           ||new.OrdID||','
           ||''''||new.DateOrd||''''||','
           ||''''||new.CustID||''''||','
           ||''''||new.ItemID||''''||','
           ||'--'||','
           ||new.Qty||','
           ||''''||new.State||''''||')') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTORDER_INS1') $;$
   -- -----------------------------------------------------
   --
   insert into V_CORDER(Price,InvID)
      select Price,(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
      from   ITEM where ItemID = new.ItemID $;$
   --
   -- Adjust ITEM quantities   --
   --
   update ITEM
   set   QonHand = QonHand - new.Qty
   where ItemID = new.ItemID $;$
   --
   -- Update customer order    --
   --
   update CUSTORDER
   set   Price = (select Price from V_CORDER)
   where OrdID = new.OrdID $;$
   --
   -- Create an invoice (default State value only):
   -- 
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,
              (select Price from V_CORDER),new.Qty,'recorded') $;$
   -- 
   -- Set invoice State to 'sent' to trigger printing of the invoice:
   --
   update CUSTINVOICE set State = 'sent' where InvID = (select InvID from V_CORDER) $;$
   -- 
   delete from V_CORDER $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTORDER_INS1') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- ====================================================== --
-- INSERT: "Quantity on hand" is NOT sufficient           --
-- ====================================================== --

create trigger TRG_CUSTORDER_INS2
after insert on CUSTORDER
for each row
when new.Qty > (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: insert into CUSTORDER ('
           ||new.OrdID||','
           ||''''||new.DateOrd||''''||','
           ||''''||new.CustID||''''||','
           ||''''||new.ItemID||''''||','
           ||'--'||','
           ||new.Qty||','
           ||''''''||')') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTORDER_INS2') $;$
   -- -----------------------------------------------------
   --
   -- Extract initial values used more than once in this trigger
   --
   insert into V_CORDER(Price,Qavail,mQeco,InvID) 
      select Price,Qavail,
             Qeco * (truncate(1.*(new.Qty - Qavail)/Qeco) + 1),
             (select coalesce(max(InvID),0)+1 from CUSTINVOICE)
      from   ITEM where ItemID = new.ItemID $;$
   --
   -- Adjust item quantities (all cases) --
   --
   update ITEM
   set   QonHand = QonHand - new.Qty
   where ItemID = new.ItemID $;$
   --
   -- ---------------------------------------------- --
   -- "Quantity available" is NOT sufficient         --
   -- ---------------------------------------------- --
   --
   -- Adjust item quantities (with supplier order):
   --
   update ITEM
   set   Qord = Qord + (select mQeco from V_CORDER)
   where ItemID = new.ItemID and new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Identify best offer if not selected yet:
   --
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)
   where ItemID = new.ItemID and SuppID is null and new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Create a supplier order
   --
   insert into SUPPORDER(OrdID,DateOrd,ItemID,SuppID,Price,Qty,State)
          select (select coalesce(max(OrdID),0) + 1 from SUPPORDER),
                  new.DateOrd,new.ItemID,SuppID,SuppPrice,
                  (select mQeco from V_CORDER),'assigned'
          from   ITEM
          where  ItemID = new.ItemID
          and    new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- ---------------------------------------------- --
   -- In both cases                                  --
   -- ---------------------------------------------- --
   --
   -- Update customer order:
   --
   update CUSTORDER
   set   Price = (select Price from V_CORDER),State = 'pending'
   where OrdID = new.OrdID $;$
   --
   -- Create a pending invoice:
   --
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,
              (select Price from V_CORDER),new.Qty,'pending') $;$
   --
   delete from V_CORDER $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTORDER_INS2') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


create trigger TRG_CUSTORDER_UPD
after update on CUSTORDER
for each row
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update CUSTORDER set '
           ||rtrim(iif(new.OrdID = old.OrdID ,   '', 'OrdID = '  ||new.OrdID ||',')
                 ||iif(new.DateOrd = old.DateOrd,'', 'DateOrd = '||''''||new.DateOrd||''''||',')
                 ||iif(new.CustID = old.CustID,  '', 'CustdID = '||''''||new.CustID ||''''||',')
                 ||iif(new.ItemID = old.ItemID,  '', 'ItemID = ' ||''''||new.ItemID ||''''||',')
                 ||iif(coalesce(new.Price,'--') = coalesce(old.Price,'--'),'','Price = '||coalesce(new.Price,'--')||',')
                 ||iif(new.Qty = old.Qty, '','Qty = '||new.Qty||',')
                 ||iif(new.State = old.State,'','State = '||''''||new.State||''''||','), ',') ) $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTORDER_UPD') $;$
   -- -----------------------------------------------------
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTORDER_UPD') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGER on ITEM                              --
-- ============================================ --

create trigger TRG_ITEM_UPD1
after update of QonHand,SuppID,SuppPrice on ITEM
for each row
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   -- Note: If no change (e.g., same "best offer"), just write "EVENT: update ITEM set"
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update ITEM set '
           ||rtrim(iif(new.QonHand = old.QonHand , '', 'QonHand = '  ||new.QonHand ||',')
                 ||iif(coalesce(new.SuppID,'--') = coalesce(old.SuppID,'--'),'','SuppID = '||coalesce(new.SuppID,'--')||',')
                 ||iif(coalesce(new.SuppPrice,'--') = coalesce(old.SuppPrice,'--'),'','SuppPrice = '||coalesce(new.SuppPrice,'--')||','), ',') ) $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_ITEM_UPD1') $;$
   -- -----------------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_ITEM_UPD1') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;

-- --------------------------------- --
-- An item has been replenished      --
-- --------------------------------- --
create trigger TRG_ITEM_UPD2
after update of Qord on ITEM
for each row
begin
   -- Only active when "Qord" has been reset to 0, that is, "item replenished"  
   --    
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Send pending customer invoices:
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update ITEM set Qord = '||new.Qord) $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_ITEM_UPD2') $;$
   -- -----------------------------------------------------
   --
   update CUSTINVOICE
   set   State = 'sent'
   where ItemID = new.ItemID and State = 'pending'
   and   new.Qord = 0 $;$
   --
   -- Reevaluate the best offer of the item:
   --
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER
                               where  ItemID = new.ItemID
                               order by Price,Delay limit 1)
   where ItemID = new.ItemID
   and   new.Qord = 0  $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_ITEM_UPD2') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGERS on CUSTINVOICE                      --
-- ============================================ --

create trigger TRG_CUSTINVOICE_INS
after insert on CUSTINVOICE
for each row
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: insert into CUSTINVOICE ('
           ||new.InvID||','
           ||''''||new.DateInv||''''||','
           ||new.OrdID||','
           ||''''||new.CustID||''''||','
           ||''''||new.ItemID||''''||','
           ||new.Price||','
           ||new.Qty||','
           ||new.Amount||','
           ||''''||new.State||''''||')') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTINVOICE_INS') $;$
   -- -----------------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTINVOICE_INS') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


create trigger TRG_CUSTINVOICE_UPD1
after update on CUSTINVOICE
for each row
when not (old.State <> 'sent' and new.State = 'sent')
and  not (old.State <> 'paid' and new.State = 'paid')
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update CUSTINVOICE set '
           ||rtrim(iif(new.InvID  = old.InvID,   '', 'InvID = '  ||new.InvID ||',')
                 ||iif(new.DateInv = old.DateInv,'', 'DateInv = '||''''||new.DateInv ||''''||',')
                 ||iif(new.OrdID  = old.OrdID,   '', 'OrdID = '  ||new.OrdID ||',')
                 ||iif(new.CustID = old.CustID,  '', 'CustdID = '||''''||new.CustID ||''''||',')
                 ||iif(new.ItemID = old.ItemID,  '', 'ItemID = ' ||''''||new.ItemID ||''''||',')
                 ||iif(new.Price  = old.Price,   '', 'Price = '  ||new.Price ||',')
                 ||iif(new.Qty = old.Qty,        '', 'Qty = '||new.Qty||',')
                 ||iif(new.Amount = old.Amount,  '', 'Amount = '||new.Amount||',')
                 ||iif(new.State = old.State,    '', 'State = '||''''||new.State||''''||','), ',') ) $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTINVOICE_UPD1') $;$
   -- -----------------------------------------------------
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTINVOICE_UPD1') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


create trigger TRG_CUSTINVOICE_UPD2
-- multiple events not allowed in SQLite: "insert" replaced by "update" --
after update of State on CUSTINVOICE
for each row
when new.State = 'sent'
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update CUSTINVOICE set State = '||''''||new.State||'''') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTINVOICE_UPD2') $;$
   -- -----------------------------------------------------
   --
   $writeFile$('D:/SQLfast/Files/INVOICE-'||new.InvID||'.txt',
                   'INVOICE no '||new.InvID
                 ||'@n--------------------'
                 ||'@n CUSTOMER: '||new.CustID
                 ||'@n     Date: '||new.DateInv
                 ||'@n     Item: '||new.ItemID
                 ||'@n      Qty: '||new.Qty
                 ||'@n    Price: '||new.Price
                 ||'@n--------------------'
                 ||'@n    Total: '||thousandSep(cast(new.Qty*new.Price as char),',')
                   ) $;$
  --
  update CUSTORDER
  set   State = 'invoiced'
  where OrdID = new.OrdID $;$
  --
  update CUSTOMER
  set   Account = Account - new.Amount
  where CustID = new.CustID $;$
  --
  -- Tracing ---------------------------------------------
  $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTINVOICE_UPD2') $;$
  -- -----------------------------------------------------
  update TLEVEL set Level = Level - 1 $;$
end;


create trigger TRG_CUSTINVOICE_UPD3
after update of State on CUSTINVOICE
for each row
when new.State = 'paid'
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update CUSTINVOICE set State = '||''''||new.State||'''') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_CUSTINVOICE_UPD3') $;$
   -- -----------------------------------------------------
  --
   update CUSTORDER
   set   State = 'closed'
   where OrdID = (select OrdID from CUSTINVOICE where InvID = new.InvID) $;$
   --
   update CUSTOMER
   set   Account = Account + new.Amount
   where CustID = new.CustID $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_CUSTINVOICE_UPD3') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGERS on SUPPORDER                        --
-- ============================================ --

create trigger TRG_SUPPORDER_INS
after insert on SUPPORDER
for each row
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: insert into SUPPORDER ('
           ||new.OrdID||','
           ||''''||new.DateOrd||''''||','
           ||''''||new.ItemID||''''||','
           ||''''||new.SuppID||''''||','
           ||new.Price||','
           ||new.Qty||','
           ||''''||new.State||''''||')') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_SUPPORDER_INS') $;$
   -- -----------------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_SUPPORDER_INS') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;


-- Replenish an item --
create trigger TRG_SUPPORDER_UPD
after update of State on SUPPORDER
for each row
when new.State = 'closed'
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)
           ||'EVENT: update SUPPORDER set State = '||''''||new.State||'''') $;$
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  > Enter TRG_SUPPORDER_UPD') $;$
   -- -----------------------------------------------------
   --
   update ITEM
   set   QonHand = QonHand + new.Qty, Qord = Qord - new.Qty
   where ItemID = new.ItemID $;$
   --
   -- Tracing ---------------------------------------------
   $message$((select repeat(' ',Level*4) from TLEVEL)||'  < Exit TRG_SUPPORDER_UPD') $;$
   -- -----------------------------------------------------
   update TLEVEL set Level = Level - 1 $;$
end;

closeDB;

dropVariable ;;

return;



