createOrReplaceDB RETAIL.db;

create table CUSTOMER (
        CustID  char(10) not null,
        Name    char(32) not null,
        Address char(60) not null,
        City    char(30) not null,
        Account decimal(9,2) not null default 0.0,
        constraint PKCUST primary key (CustID));

create table ITEM (
        ItemID      char(15) not null,
        Description char(60) not null,
        Price       decimal(4,2) not null,
        SuppID      char(10),
        SuppPrice   decimal(4,2),
        QonHand     integer not null,
        Qord        integer default 0,
        Qavail      integer not null generated always as (QonHand+Qord),
        Qeco        integer not null,
        constraint PKITEM primary key (ItemID));

create table CUSTORDER (
        OrdID   integer  not null,
        DateOrd date     not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2),
        Qty     integer  not null,
        State   char(20) not null default 'recorded',
        constraint PKCORD primary key (OrdID),
        constraint FKORDCUST foreign key (CustID)
           references CUSTOMER
           on delete no action on update cascade,
        constraint FKORDITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade,
        constraint CHKQTY check(Qty > 0));

create table CUSTINVOICE (
        InvID   integer  not null,
        DateInv date     not null,
        OrdID   integer  not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Qty     integer  not null,
        Amount  integer  not null generated always as (Qty*Price),
        State   char(20) not null default 'recorded',
        constraint PKCINV primary key (InvID),
        constraint FKINVORD foreign key (OrdID)
           references CUSTORDER
           on delete no action on update cascade);

create table SUPPLIER (
        SuppID  char(10) not null,
        Name    char(32) not null,
        City    char(30) not null,
        constraint PKSUPP primary key (SuppID));

create table OFFER (
        SuppID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Delay   integer not null,
        constraint PKOFF primary key (SuppID,ItemID),
        constraint FKOFFSUPP foreign key (SuppID)
           references SUPPLIER
           on delete no action on update cascade,
        constraint FKOFFITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade);

create table SUPPORDER (
        OrdID   integer not null,
        DateOrd date not null,
        ItemID  char(15) not null,
        SuppID  char(10),
        Price   decimal(4,2),
        Qty     integer  not null,
        State  varchar(20) not null default 'recorded',
        constraint PKSORD primary key (OrdID)
        constraint FKORDSUPP foreign key (SuppID,ItemID)
           references OFFER
           on delete no action on update cascade);

-- Variables used in CUSTORDER processing
create table V_CORDER as
    select Price,Qavail,0 as mQeco,0 as InvID
    from   ITEM where False;
create table TLEVEL as select 0 as Level;;

set ; = ;;
set trace     = select writeTrace;
set writeFile = select writeFile;


-- CUSTORDER
-- create trigger TRG_CUSTORDER_INS1
-- create trigger TRG_CUSTORDER_INS2

-- ITEM
-- create trigger TRG_ITEM_UPD

-- CUSTINVOICE
-- create trigger TRG_CUSTINVOICE_UPD2
-- create trigger TRG_CUSTINVOICE_UPD3

-- SUPPORDER
-- create trigger TRG_SUPPORDER_UPD



-- ====================================================== --
-- TRIGGERS on CUSTORDER                                  --
-- ====================================================== --

-- ---------------------------------------------- --
-- INSERT: "Quantity on hand" is sufficient       --
-- ---------------------------------------------- --

create trigger TRG_CUSTORDER_INS1
after insert on CUSTORDER
for each row
when new.Qty <= (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Extract initial values used more than once in this trigger
   --
   -- Tracing ---------------------------------------------
   $trace$(2,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
          'insert into V_CORDER(Price,InvID) select Price, (select coalesce(max(InvID),0)+1 from CUSTINVOICE) from ITEM where ItemID = new.ItemID',
          'insert into V_CORDER(Price,InvID) select ' || (select Price||', '||(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
                                                          from   ITEM
                                                          where ItemID = new.ItemID) || ' from ITEM where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
   insert into V_CORDER(Price,InvID)
          select Price,(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
          from   ITEM where ItemID = new.ItemID $;$
   --
   -- Adjust ITEM quantities   --
   --
   -- Tracing ---------------------------------------------
   $trace$(5,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
          'update ITEM set QonHand = QonHand - new.Qty where ItemID = new.ItemID',
          'update ITEM set QonHand = QonHand - ' || new.Qty||' where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
   update ITEM set QonHand = QonHand - new.Qty where ItemID = new.ItemID $;$
   --
   -- Update customer order    --
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
          'update CUSTORDER set Price = (select Price from V_CORDER) where OrdID = new.OrdID ',
          'update CUSTORDER set Price = ' || (select Price from V_CORDER) || ' where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
   update CUSTORDER set Price = (select Price from V_CORDER) where OrdID = new.OrdID $;$
   --
   -- Create an invoice (default State value only):
   --
   -- Tracing ---------------------------------------------
   $trace$(7,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
          'insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State) '
        ||'values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,'
        ||'(select Price from V_CORDER),new.Qty,''recorded'')',
        --
          'insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State) '
        ||'values ('||(select InvID from V_CORDER)||', '''||new.DateOrd||''', '||new.OrdID||', '''||new.CustID||''', '''||new.ItemID||''', '
        ||(select Price from V_CORDER)||', '||new.Qty||', ''recorded'')',1) $;$
   -- -----------------------------------------------------
   -- 
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,
              (select Price from V_CORDER),new.Qty,'recorded') $;$
   -- 
   -- Set invoice State to 'sent' to trigger printing of the invoice:
   --
   -- Tracing ---------------------------------------------
   $trace$(10,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
           'update CUSTINVOICE set State = ''sent'' where InvID = (select InvID from V_CORDER)',
           'update CUSTINVOICE set State = ''sent'' where InvID = '||(select InvID from V_CORDER),1) $;$
   -- -----------------------------------------------------
   --
   update CUSTINVOICE set State = 'sent' where InvID = (select InvID from V_CORDER) $;$
   --
   -- Tracing ---------------------------------------------
   $trace$(11,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS1',(select Level from TLEVEL),0,
           'delete from V_CORDER','delete from V_CORDER',1) $;$
   -- -----------------------------------------------------
   -- 
   delete from V_CORDER $;$
   --
   update TLEVEL set Level = Level - 1 $;$
end;


-- ---------------------------------------------- --
-- "Quantity on hand" is NOT sufficient           --
-- ---------------------------------------------- --

create trigger TRG_CUSTORDER_INS2
after insert on CUSTORDER
for each row
when new.Qty > (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Extract initial values used more than once in this trigger
   --
   $trace$(2,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'insert into V_CORDER(Price,Qavail,mQeco,InvID) '
        ||'select Price,Qavail,'
        ||'Qeco * (truncate(1.*(new.Qty - Qavail)/Qeco) + 1),'
        ||'(select coalesce(max(InvID),0)+1 from CUSTINVOICE) from ITEM where ItemID = new.ItemID',
        --
          'insert into V_CORDER(Price,Qavail,mQeco,InvID) '
        ||'select '
        ||(select Price||', '||Qavail||', '||(Qeco*(truncate(1.*(new.Qty-Qavail)/Qeco)+1)) from ITEM where ItemID = new.ItemID)||', '
        ||(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
        ||' from ITEM where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
   insert into V_CORDER(Price,Qavail,mQeco,InvID) 
          select Price,Qavail,Qeco * (truncate(1.*(new.Qty - Qavail)/Qeco) + 1),(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
          from   ITEM where ItemID = new.ItemID $;$
   --
   -- Adjust item quantities (all cases) --
   --
   -- Tracing ---------------------------------------------
   $trace$(5,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'update ITEM set QonHand = QonHand - new.Qty where ItemID = new.ItemID',
          'update ITEM set QonHand = QonHand - ' || new.Qty||' where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
  update ITEM set QonHand = QonHand - new.Qty where ItemID = new.ItemID $;$
   --
   -- ---------------------------------------------- --
   -- "Quantity available" is NOT sufficient         --
   -- ---------------------------------------------- --
   --
   -- Adjust item quantities (with supplier order):
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'update ITEM set Qord = Qord + (select mQeco from V_CORDER) where ItemID = new.ItemID and new.Qty > (select Qavail from V_CORDER)',
          'update ITEM set Qord = Qord + '||(select mQeco from V_CORDER)||' where ItemID = ''' ||new.ItemID|| ''''||' and '||new.Qty||' > '||(select Qavail from V_CORDER),
          new.Qty > (select Qavail from V_CORDER)) $;$
   -- -----------------------------------------------------
   --
   update ITEM
   set   Qord = Qord + (select mQeco from V_CORDER) where ItemID = new.ItemID and new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Identify best offer if not selected yet:
   --
   -- Tracing ---------------------------------------------
   $trace$(7,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'update ITEM set (SuppID,SuppPrice) = (select SuppID,Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)'
        ||' where ItemID = new.ItemID and SuppID is null and new.Qty > (select Qavail from V_CORDER)',
        --
          'update ITEM set (SuppID,SuppPrice) = ('||(select SuppID||', '||Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)
        ||') where ItemID = '''||new.ItemID||''' and SuppID is null and '||new.Qty||' > '||(select Qavail from V_CORDER),
          (select ((SuppID is null) and (new.Qty > (select Qavail from V_CORDER))) from ITEM where ItemID = new.ItemID)) $;$
   -- -----------------------------------------------------
   --
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)
   where ItemID = new.ItemID and SuppID is null and new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Create a supplier order
   --
   -- Tracing ---------------------------------------------
   $trace$(10,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'insert into SUPPORDER(OrdID,DateOrd,ItemID,SuppID,Price,Qty,State) '
        ||'select (select coalesce(max(OrdID),0) + 1 from SUPPORDER),'
        ||        'new.DateOrd,new.ItemID,SuppID,SuppPrice,(select mQeco from V_CORDER),''assigned'''
        ||' from ITEM'
        ||' where ItemID = new.ItemID'
        ||' and new.Qty > (select Qavail from V_CORDER)',
        --
          'insert into SUPPORDER(OrdID,DateOrd,ItemID,SuppID,Price,Qty,State) '
        ||'select '||(select coalesce(max(OrdID),0) + 1 from SUPPORDER)||', '''
        || new.DateOrd||''', '''|| new.ItemID||''', '''
        ||(select SuppID||''', '||SuppPrice from ITEM where ItemID = new.ItemID)||', '
        ||(select mQeco from V_CORDER)||', ' || '''assigned'''
        ||' from ITEM where ItemID = ''' ||new.ItemID|| ''''
        ||' and new.Qty > ' || (select Qavail from V_CORDER),
         (new.Qty > (select Qavail from V_CORDER)) 
          ) $;$
   -- -----------------------------------------------------
   --
   insert into SUPPORDER(OrdID,DateOrd,ItemID,SuppID,Price,Qty,State)
          select (select coalesce(max(OrdID),0) + 1 from SUPPORDER),
                  new.DateOrd,new.ItemID,SuppID,SuppPrice,
                  (select mQeco from V_CORDER),'assigned'
          from   ITEM
          where  ItemID = new.ItemID
          and    new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- ---------------------------------------------- --
   -- In both cases                                  --
   -- ---------------------------------------------- --
   --
   -- Update customer order:
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'update CUSTORDER set Price = (select Price from V_CORDER), State = ''pending'' where OrdID = new.OrdID ',
          'update CUSTORDER set Price = ' || (select Price from V_CORDER) || ', State = ''pending''' 
        ||' where ItemID = ''' ||new.ItemID|| '''',1) $;$
   -- -----------------------------------------------------
   --
   update CUSTORDER
   set   Price = (select Price from V_CORDER),State = 'pending'
   where OrdID = new.OrdID $;$
   --
   -- Create a pending invoice:
   --
   -- Tracing ---------------------------------------------
   $trace$(7,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
          'insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State) '
        ||'values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,'
        ||'(select Price from V_CORDER),new.Qty,''pending'')',
        --
          'insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State) '
        ||'values ('||(select InvID from V_CORDER)||', '''||new.DateOrd||''', '||new.OrdID||', '''||new.CustID||''', '''||new.ItemID||''', '
        ||(select Price from V_CORDER)||', '||new.Qty||', ''pending'')',1) $;$
   -- -----------------------------------------------------
   --
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),new.DateOrd,new.OrdID,new.CustID,new.ItemID,
              (select Price from V_CORDER),new.Qty,'pending') $;$
   --
   -- Tracing ---------------------------------------------
   $trace$(11,current_timestamp_full(),'RETAIL.db/TRG_CUSTORDER_INS2',(select Level from TLEVEL),0,
           'delete from V_CORDER','delete from V_CORDER',1) $;$
   -- -----------------------------------------------------
   --
   delete from V_CORDER $;$
   --
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGER on ITEM                              --
-- ============================================ --

-- --------------------------------- --
-- An item has been replenished      --
-- --------------------------------- --
create trigger TRG_ITEM_UPD
after update of Qord on ITEM
for each row
when old.Qord > 0 and new.Qord = 0
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Send pending customer invoices:
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_ITEM_UPD2',(select Level from TLEVEL),0,
          'update CUSTINVOICE set State = ''sent'' where ItemID = new.ItemID and State = ''pending''',
          'update CUSTINVOICE set State = ''sent'' where ItemID = '''||new.ItemID||''' and State = ''pending''',1) $;$
   -- -----------------------------------------------------
   --
   update CUSTINVOICE
   set   State = 'sent'
   where ItemID = new.ItemID and State = 'pending' $;$
   --
   -- Reevaluate the best offer of the item:
   --
   -- Tracing ---------------------------------------------
   $trace$(7,current_timestamp_full(),'RETAIL.db/TRG_ITEM_UPD2',(select Level from TLEVEL),0,
          'update ITEM set (SuppID,SuppPrice) = (select SuppID,Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)'
        ||' where ItemID = new.ItemID',
        --
          'update ITEM set (SuppID,SuppPrice) = ('||(select SuppID||', '||Price from OFFER where ItemID = new.ItemID order by Price,Delay limit 1)
        ||') where ItemID = '''||new.ItemID||'''',1) $;$
   -- -----------------------------------------------------
   --
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER
                               where  ItemID = new.ItemID
                               order by Price,Delay limit 1)
   where ItemID = new.ItemID $;$
   --
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGERS on CUSTINVOICE                      --
-- ============================================ --


create trigger TRG_CUSTINVOICE_UPD2
-- multiple events not allowed in SQLite: "insert" replaced by "update" --
after update of State on CUSTINVOICE
for each row
when new.State = 'sent'
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTINVOICE_UPD2',(select Level from TLEVEL),0,
          'writeFile(''D:/SQLfast/Files/INVOICE-''||new.InvID||.txt'' ...)',
          'writeFile(''D:/SQLfast/Files/INVOICE-'||new.InvID||'.txt'' ...)' ,1) $;$
   -- -----------------------------------------------------
   --
   $writeFile$('D:/SQLfast/Files/INVOICE-'||new.InvID||'.txt',
                   'INVOICE no '||new.InvID
                 ||'@n--------------------'
                 ||'@n CUSTOMER: '||new.CustID
                 ||'@n     Date: '||new.DateInv
                 ||'@n     Item: '||new.ItemID
                 ||'@n      Qty: '||new.Qty
                 ||'@n    Price: '||new.Price
                 ||'@n--------------------'
                 ||'@n    Total: '||thousandSep(cast(new.Qty*new.Price as char),',')
                   ) $;$
  --
  -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTINVOICE_UPD2',(select Level from TLEVEL),0,
          'update CUSTORDER set State = ''invoiced'' where OrdID = new.OrdID',
          'update CUSTORDER set State = ''invoiced'' where OrdID = '||new.OrdID,1) $;$
  -- -----------------------------------------------------
  --
  update CUSTORDER set State = 'invoiced' where OrdID = new.OrdID $;$
  --
  -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTINVOICE_UPD2',(select Level from TLEVEL),0,
          'update CUSTOMER set Account = Account - new.Amount where CustID = new.CustID',
          'update CUSTOMER set Account = Account - '||new.Amount||' where CustID = '''||new.CustID||'''',1) $;$
  -- -----------------------------------------------------
  --
  update CUSTOMER set Account = Account - new.Amount where CustID = new.CustID $;$
  --
  update TLEVEL set Level = Level - 1 $;$
end;


create trigger TRG_CUSTINVOICE_UPD3
after update of State on CUSTINVOICE
for each row
when new.State = 'paid'
begin
   update TLEVEL set Level = Level + 1 $;$
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTINVOICE_UPD3',(select Level from TLEVEL),0,
          'update CUSTORDER set State = ''closed'' where OrdID = (select OrdID from CUSTINVOICE where InvID = new.InvID)',
          'update CUSTORDER set State = ''closed'' where OrdID = '||(select OrdID from CUSTINVOICE where InvID = new.InvID),1) $;$
  -- -----------------------------------------------------
  --
   update CUSTORDER set State = 'closed' where OrdID = (select OrdID from CUSTINVOICE where InvID = new.InvID) $;$
   --
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_CUSTINVOICE_UPD3',(select Level from TLEVEL),0,
          'update CUSTOMER set Account = Account + new.Amount where CustID = new.CustID',
          'update CUSTOMER set Account = Account + '||new.Amount||' where CustID = '''||new.CustID||'''',1) $;$
   -- -----------------------------------------------------
   --
   update CUSTOMER set Account = Account + new.Amount where CustID = new.CustID $;$
   --
   update TLEVEL set Level = Level - 1 $;$
end;


-- ============================================ --
-- TRIGGERS on SUPPORDER                        --
-- ============================================ --

-- Replenish an item --
create trigger TRG_SUPPORDER_UPD
after update of State on SUPPORDER
for each row
when new.State = 'closed'
begin
   update TLEVEL set Level = Level + 1 $;$
   -- Tracing ---------------------------------------------
   $trace$(6,current_timestamp_full(),'RETAIL.db/TRG_SUPPORDER_INS',(select Level from TLEVEL),0,
          'update ITEM set QonHand = QonHand + new.Qty,Qord = Qord - new.Qty where ItemID = new.ItemID',
          'update ITEM set QonHand = QonHand + '||new.Qty||', Qord = Qord - '||new.Qty||' where ItemID = '''||new.ItemID||'''',1) $;$
   -- -----------------------------------------------------
   --
   update ITEM
   set   QonHand = QonHand + new.Qty, Qord = Qord - new.Qty where ItemID = new.ItemID $;$
   --
   update TLEVEL set Level = Level - 1 $;$
end;

closeDB;

dropVariable ;;

return;



