createOrReplaceDB RETAIL.db;

create table CUSTOMER (
        CustID  char(10) not null,
        Name    char(32) not null,
        Address char(60) not null,
        City    char(30) not null,
        Account decimal(9,2) not null default 0.0,
        constraint PKCUST primary key (CustID));

create table ITEM (
        ItemID      char(15) not null,
        Description char(60) not null,
        Price       decimal(4,2) not null,
        SuppID      char(10),
        SuppPrice   decimal(4,2),
        QonHand     integer not null,
        Qord        integer default 0,
        Qavail      integer not null generated always as (QonHand+Qord),
        Qeco        integer not null,
        constraint PKITEM primary key (ItemID));

create table CUSTORDER (
        OrdID   integer  not null,
        DateOrd date     not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2),
        Qty     integer  not null,
        State   char(20) not null default 'recorded',
        constraint PKCORD primary key (OrdID),
        constraint FKORDCUST foreign key (CustID)
           references CUSTOMER
           on delete no action on update cascade,
        constraint FKORDITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade,
        constraint CHKQTY check(Qty > 0));

create table CUSTINVOICE (
        InvID   integer  not null,
        DateInv date     not null,
        OrdID   integer  not null,
        CustID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Qty     integer  not null,
        Amount  integer  not null generated always as (Qty*Price),
        State   char(20) not null default 'recorded',
        constraint PKCINV primary key (InvID),
        constraint FKINVORD foreign key (OrdID)
           references CUSTORDER
           on delete no action on update cascade);

create table SUPPLIER (
        SuppID  char(10) not null,
        Name    char(32) not null,
        City    char(30) not null,
        constraint PKSUPP primary key (SuppID));

create table OFFER (
        SuppID  char(10) not null,
        ItemID  char(15) not null,
        Price   decimal(4,2) not null,
        Delay   integer not null,
        constraint PKOFF primary key (SuppID,ItemID),
        constraint FKOFFSUPP foreign key (SuppID)
           references SUPPLIER
           on delete no action on update cascade,
        constraint FKOFFITEM foreign key (ItemID)
           references ITEM
           on delete no action on update cascade);

create table SUPPORDER (
        OrdID   integer not null,
        DateOrd date not null,
        ItemID  char(15) not null,
        SuppID  char(10),
        Price   decimal(4,2),
        Qty     integer  not null,
        State  varchar(20) not null default 'recorded',
        constraint PKSORD primary key (OrdID)
        constraint FKORDSUPP foreign key (SuppID,ItemID)
           references OFFER
           on delete no action on update cascade);

-- Variables used in CUSTORDER processing
create table V_CORDER as
    select Price,Qavail,0 as mQeco,0 as InvID
    from   ITEM where False;

set ; = ;;

-- CUSTORDER
-- create trigger TRG_CUSTORDER_INS1
-- create trigger TRG_CUSTORDER_INS2

-- ITEM
-- create trigger TRG_ITEM_UPD

-- CUSTINVOICE
-- create trigger TRG_CUSTINVOICE_UPD2
-- create trigger TRG_CUSTINVOICE_UPD3

-- SUPPORDER
-- create trigger TRG_SUPPORDER_UPD


-- ============================================ --
-- TRIGGERS on CUSTORDER                        --
-- ============================================ --

-- ---------------------------------------------- --
-- INSERT: "Quantity on hand" is sufficient       --
-- ---------------------------------------------- --

create trigger TRG_CUSTORDER_INS1
after insert on CUSTORDER
for each row
when new.Qty <= (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   --
   -- Extract initial values used more than once in this trigger
   insert into V_CORDER(Price,InvID)
          select Price,(select coalesce(max(InvID),0)+1 from CUSTINVOICE)
          from   ITEM where ItemID = new.ItemID $;$
   -- 
   -- Adjust ITEM quantities   --
   update ITEM
   set   QonHand = QonHand - new.Qty
   where ItemID = new.ItemID $;$
   --
   -- Update customer order    --
   update CUSTORDER
   set    Price = (select Price from V_CORDER)
   where  OrdID = new.OrdID $;$
   --
   -- Create an invoice (no State yet):
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),
               new.DateOrd,
               new.OrdID,
               new.CustID,
               new.ItemID,
              (select Price from V_CORDER),
               new.Qty,
               '') $;$
   --
   -- Set invoice State to 'valid' to trigger printing of the invoice:
   update CUSTINVOICE
   set    State = 'sent'
   where  InvID = (select InvID from V_CORDER) $;$
   --
   delete from V_CORDER $;$
   --
end;

-- ---------------------------------------------- --
-- "Quantity on hand" is NOT sufficient           --
-- ---------------------------------------------- --

create trigger TRG_CUSTORDER_INS2
after insert on CUSTORDER
for each row
when new.Qty > (select QonHand from ITEM where ItemID = new.ItemID)
and  not exists (select * from CUSTINVOICE where OrdID = new.OrdID)
begin
   --
   -- Extract initial values used more than once in this trigger
   insert into V_CORDER(Price,Qavail,mQeco,InvID) 
          select Price,Qavail,
                 Qeco * (truncate(1.*(new.Qty - Qavail)/Qeco) + 1),
                 (select coalesce(max(InvID),0)+1 from CUSTINVOICE)
          from   ITEM where ItemID = new.ItemID $;$
   --
   -- Adjust item quantities --;
   update ITEM
   set   QonHand = QonHand - new.Qty
   where ItemID = new.ItemID $;$
   --
   -- ---------------------------------------------- --
   -- "Quantity available" is NOT sufficient         --
   -- ---------------------------------------------- --
   --
   -- Adjust item quantities (with supplier order):
   update ITEM
   set   Qord   = Qord + (select mQeco from V_CORDER)
   where ItemID = new.ItemID
   and   new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Identify best offer if not selected yet:
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER
                               where  ItemID = new.ItemID
                               order by Price,Delay limit 1)
   where ItemID = new.ItemID
   and   SuppID is null 
   and   new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- Create a supplier order
   insert into SUPPORDER(OrdID,DateOrd,ItemID,SuppID,Price,Qty,State)
          select (select coalesce(max(OrdID),0) + 1 from SUPPORDER),
                  new.DateOrd,
                  new.ItemID,
                  SuppID,
                  SuppPrice,
                  (select mQeco from V_CORDER),
                  'assigned'
          from   ITEM
          where  ItemID = new.ItemID
          and    new.Qty > (select Qavail from V_CORDER) $;$
   --
   -- ---------------------------------------------- --
   -- In both cases                                  --
   -- ---------------------------------------------- --
   --
   -- Update customer order:
   update CUSTORDER
   set   Price = (select Price from V_CORDER),
         State = 'pending'
   where OrdID = new.OrdID $;$
   --
   -- Create a pending invoice:
   insert into CUSTINVOICE(InvID,DateInv,OrdID,CustID,ItemID,Price,Qty,State)
      values ((select InvID from V_CORDER),
              new.DateOrd,
              new.OrdID,
              new.CustID,
              new.ItemID,
              (select Price from V_CORDER),
              new.Qty,
              'pending') $;$
   --
   delete from V_CORDER $;$
   --
end;


-- ============================================ --
-- TRIGGER on ITEM                              --
-- ============================================ --

-- --------------------------------- --
-- When an item has been replenished --
-- --------------------------------- --

create trigger TRG_ITEM_UPD
after update of Qord on ITEM
for each row
when new.Qord = 0
begin
   --
   -- Send pending customer invoices:
   update CUSTINVOICE
   set   State = 'sent'
   where ItemID = new.ItemID
   and   State = 'pending' $;$
   --
   -- Reevaluate the best offer of the item:
   update ITEM
   set   (SuppID,SuppPrice) = (select SuppID,Price from OFFER
                               where  ItemID = new.ItemID
                               order by Price,Delay limit 1)
   where ItemID = new.ItemID $;$
   --
end;


-- ============================================ --
-- TRIGGERS on CUSTINVOICE                      --
-- ============================================ --

create trigger TRG_CUSTINVOICE_UPD1
-- multiple events not allowed in SQLite: "insert" replaced by "update" --
after update of State on CUSTINVOICE
for each row
when new.State = 'sent'
begin
   --
   select raise(IGNORE)
   where writeFile('D:/SQLfast/Files/INVOICE-'||new.InvID||'.txt',
                   'INVOICE no '||new.InvID
                 ||'@n--------------------'
                 ||'@n CUSTOMER: '||new.CustID
                 ||'@n     Date: '||new.DateInv
                 ||'@n     Item: '||new.ItemID
                 ||'@n      Qty: '||new.Qty
                 ||'@n    Price: '||new.Price
                 ||'@n--------------------'
                 ||'@n    Total: '||thousandSep(cast(new.Qty*new.Price as char),',')
                   ) $;$
  --
  update CUSTORDER set State = 'invoiced' where OrdID = new.OrdID $;$
  --
  update CUSTOMER set Account = Account - new.Amount where CustID = new.CustID $;$
  --
end;


create trigger TRG_CUSTINVOICE_UPD2
after update of State on CUSTINVOICE
for each row
when new.State = 'paid'
begin
   --
   update CUSTORDER
   set   State = 'closed'
   where OrdID = (select OrdID from CUSTINVOICE where InvID = new.InvID) $;$
   --
   update CUSTOMER
   set   Account = Account + new.Amount
   where CustID = new.CustID $;$
   --
end;


-- ============================================ --
-- TRIGGERS on SUPPORDER                        --
-- ============================================ --

-- Replenish an item --
create trigger TRG_SUPPORDER_UPD
after update of State on SUPPORDER
for each row
when new.State = 'closed'
begin
   --
   update ITEM
   set   QonHand = QonHand + new.Qty,
         Qord    = Qord - new.Qty
   where ItemID = new.ItemID $;$
   --
end;

closeDB;

dropVariable ;;
return;











