-- ------------------------------------------------------- --
-- Analyze trigger circuits                                --
-- ------------------------------------------------------- --
-- July 2020 --

write-b -- Analyze the graph the triggers of database "$current-db$" --;

openDB Dictionary-of-$current-db$;

-- Complete column TrigCol in SYS_TRIGGER_FULL for general "update" --
-- (general "update" = with no column list)                         --
-- ---------------------------------------------------------------- --

update SYS_TRIGGER_FULL
set    TrigCol = (select group_concat(ColName,',') from SYS_COLUMN
                  where  TableID = SYS_TRIGGER_FULL.TableID)
where  TrigEvent = 'update' and TrigCol = '';


-- Create table BINARY(<trigger-1> may-fires <trigger-2>)  --
-- ------------------------------------------------------- --

write-b -- The trigger binary graph (table BINARY) --;
create temp table BINARY
as select TrigID1,TrigID2
   from (select
                -- The firing trigger ...
                C1.TrigID as TrigID1,C1.StatOper,C1.StatCol,
                -- ... and the trigger(s) it may fire
                T2.TrigID as TrigID2,T2.TrigCol
         from   SYS_TRIGGER_COMP_FULL C1,
                SYS_TRIGGER_FULL T2
         where  C1.StatTable = T2.TableName
         and    C1.StatOper = T2.TrigEvent
         and    case when C1.StatOper = 'update'
                     then iif(itemInter(C1.StatCol,T2.TrigCol,',') = '',False,True)
                     else True
                end);

select TrigID1,'may fire' as "may fires",TrigID2 from BINARY;

write-b -- Identifying the trigger circuit(s) (view CIRCUIT) --;
create temp view CIRCUIT(Circuit) 
as with PATHS(TFrom,TTo,Path) as
     (select TrigID1,TrigID2,'.'||TrigID1||'.'||TrigID2||'.' from BINARY
         union
      select TFrom,TrigID2,Path||TrigID2||'.'
      from   PATHS,BINARY
      where  TTo = TrigID1 and Path not like '.%.'||TrigID2||'.%'
      and    length(Path) < 18
     )
select Path from PATHS where TFrom = TTo;

select * from CIRCUIT;

extract ncirc = select count(*) from CIRCUIT;
if ($ncirc$ = 0);
   write-b -- --------------------------------------------------- --
         @n-- Diagnosis: The trigger graph is acyclic.            --
         @n--            There is no risk of infinite execution.  --
         @n-- --------------------------------------------------- --;
   goto SKIP;
endif;

write-b -- By grouping permutations, we keep unique circuit(s) --;
create temp view UNIQUE_CIRCUIT 
as  select trim(min(Circuit),'.') as Circuits from CIRCUIT group by itemSort(trim(Circuit,'.'),0,'1','.');

select * from UNIQUE_CIRCUIT;

extract ncirc = select count(*) from UNIQUE_CIRCUIT;

write-b -- ----------------------------------------------------------- --
      @n-- Diagnosis: the trigger graph includes $ncirc$ unique circuits     --
      @n-- Each of them must be carefully examined and tested          --
      @n-- to make sure that it will never lead to infinite execution. --
      @n-- ----------------------------------------------------------- --@n ;

label SKIP;

write-b -- If we only want to know whether the graph includes circuit(s) --
      @n   without identifying them, we can proceed by pruning the trigger graph.;

write-b -- Let us examine the trigger binary graph (table BINARY) --
      @n   and its successive reductions.;

write-b Initial state of table BINARY:;
select * from BINARY;
extract N1 = select count(*) from BINARY;
write $N1$ arcs remaining;
set N2 = 999;
while ($N1$ < $N2$);
   delete from BINARY where TrigID2 not in (select TrigID1 from BINARY);
   write-b Outgoing arcs deleted:;
   select * from BINARY;
   extract N1 = select count(*) from BINARY;
   write $N1$ arcs remaining.;
   delete from BINARY where TrigID1 not in (select TrigID2 from BINARY);
   write-b Ingoing arcs deleted:;
   select * from BINARY;
   set N2 = $N1$;
   extract N1 = select count(*) from BINARY;
   write $N1$ arcs remaining.;
endwhile;

if ($N1$ = 0);
   write-b -- ----------------------------------------------------------------- --
         @n-- Diagnosis: the schema of "$current-db$" includes no circuit
         @n-- ----------------------------------------------------------------- --;
else;
   write-b -- ----------------------------------------------------------------- --
         @n-- Diagnosis: the schema of "$current-db$" includes circuit(s)
         @n--            but we don't know how many.
         @n-- ----------------------------------------------------------------- --;

endif;

abortDB;

closeDB;

write-b End of script;
return;