-- Database schema for valid time with granularity "date" --
-- Event times (either start or end) are provided by the user.
-- Events are reported in table LOG --


createOrReplaceDB TDB.db;

write-b -- ------------------------------------------- --;
write   -- CREATE VALID-TIME TEMPORAL DATABASE         --;
write   --  Small version: 3 projects, 8 employees     --;
write-a -- ------------------------------------------- --;

-- Parameters --
-- ---------- --
-- timestamp of future
   set Tfuture = 9999-12-31;

set errors = 0;
set onerror = 1;
askCombo [/bAbout loading errors] 
         [selectMany-u errors = generate errors]
        |[selectOne-u onerror = [/bIn case of error] stop execution|continue execution];
  if ('$DIALOGbutton$' = 'Cancel') stop;


create table H_PROJECT(
   TITLE    char(20)   not null,
   start    date       not null,
   end      date       not null default '$Tfuture$',
   THEME    char(22)   not null,
   BUDGET   decimal(8) not null,
   primary key (TITLE,start));

create table H_EMPLOYEE(
   CODE     char(5)  not null,
   start    date     not null,
   end      date     not null default '$Tfuture$',
   NAME     char(10) not null,
   STATUS   char(1)  not null,
   CITY     char(10) not null,
   PROJECT  char(20) not null,
   primary key (CODE,start));

create table LOG (
   SEQ   integer not null primary key autoincrement,
   OPER  varchar(10),
   TYPE  varchar(24),
   START date,
   ID    varchar(24),
   MESSAGE varchar(256),
   DIAG char(6) default 'FAILED');


-- Current projects --
-- --------------- --
create view PROJECT(TITLE,start,end,THEME,BUDGET)
as select TITLE,start,end,THEME,BUDGET
   from   H_PROJECT
   where  end = '$Tfuture$';

-- Current employees --
-- ---------------- --
create view EMPLOYEE(CODE,start,end,NAME,STATUS,CITY,PROJECT)
as select CODE,start,end,NAME,STATUS,CITY,PROJECT
   from   H_EMPLOYEE
   where  end = '$Tfuture$';


-- Allow ';' at the end of a query in the body of a trigger --
set ; = ;;

-- ------------------ --
-- PROJECT MANAGEMENT --
-- ------------------ --

-- Start a project --
-- --------------- --
create trigger TRG_PRO_INSERT instead of insert on PROJECT
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_PROJECT where TITLE = new.TITLE) $;$
-- 2. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start) $;$
-- 3. insert data in H_PROJECT --
   insert into H_PROJECT(TITLE,start,THEME,BUDGET) 
   values (new.TITLE,new.start,new.THEME,new.BUDGET) $;$
-- 4. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('new','PROJECT',new.start,new.TITLE,
                 new.THEME||', '||cast(new.BUDGET as char),'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;

-- Update a project --
-- ---------------- --
create trigger TRG_PRO_UPDATE instead of update on PROJECT
when new.end = '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.THEME = old.THEME and new.BUDGET = old.BUDGET $;$
-- 3. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start)
   -- also, must be "monotonic"
   or     new.start <= (select start from H_PROJECT
                        where  TITLE = old.TITLE
                        and    end = '$Tfuture$') $;$
-- 4. close the current state --
   update H_PROJECT
   set    end = new.start 
   where  TITLE = old.TITLE and end = '$Tfuture$' $;$
-- 5. insert new current state in H_PROJECT (TITLE preserved) --
   insert into H_PROJECT(TITLE,start,THEME,BUDGET) 
   values (old.TITLE,new.start,new.THEME,new.BUDGET) $;$
-- 6. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('update','PROJECT',new.start,old.TITLE,
                 new.THEME||', '||cast(new.BUDGET as char),'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;

-- Stop a project --
-- -------------- --
create trigger TRG_PRO_DELETE instead of update on PROJECT
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from H_PROJECT
                      where  TITLE = old.TITLE
                      and    end = '$Tfuture$') $;$
-- 3. No more employees on this project --
   select raise(ABORT,'dependent EMPLOYEEs still exist')
   where  exists(select * from H_EMPLOYEE
                 where  PROJECT = old.TITLE and end = '$Tfuture$') $;$
-- 4. close the current state --
   update H_PROJECT
   set    end = new.end 
   where  TITLE = old.TITLE and end = '$Tfuture$' $;$
-- 5. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('delete','PROJECT',new.end,old.TITLE,null,'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;


-- ------------------- --
-- EMPLOYEE MANAGEMENT --
-- ------------------- --

-- Create an employee --
-- ------------------ --
create trigger TRG_EMP_INSERT instead of insert on EMPLOYEE
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from H_EMPLOYEE where CODE = new.CODE) $;$
-- 2. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start) $;$
-- 3. The target PROJECT must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from H_PROJECT
                     where  TITLE = new.PROJECT
                     and    end = '$Tfuture$') $;$
-- 4. insert data in H_EMPLOYEE with the last time ID --
   insert into H_EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) 
   values (new.CODE,new.start,new.NAME,new.STATUS,new.CITY,new.PROJECT) $;$
-- 5. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('new','EMPLOYEE',new.start,new.CODE,
                 new.NAME||','||new.STATUS||', '||new.CITY||', '||new.PROJECT,'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;

-- Update an employee --
-- ------------------ --
create trigger TRG_EMP_UPDATE instead of update on EMPLOYEE
when new.end = '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. Some Entity attributes must be changed --
   select raise(ABORT,'no attribute modified')
   where  new.NAME = old.NAME and new.STATUS = old.STATUS
   and    new.CITY = old.CITY and new.PROJECT = old.PROJECT $;$
-- 3. "new.start" is a valid date --
   select raise(ABORT,'invalid start date')
   where  isNotDate(new.start)
   -- also, must be "monotonic"
   or     new.start <= (select start from H_EMPLOYEE
                        where  CODE = old.CODE
                        and    end = '$Tfuture$') $;$
-- 4. The new target PROJECT must exist since "new.start" --
   select raise(ABORT,'target PROJECT does not exist')
   where  new.PROJECT <> old.PROJECT
   and    not (    exists(select * from H_PROJECT
                          where  TITLE = new.PROJECT
                          and    end = '$Tfuture$')
               and new.start >= (select min(start) from H_PROJECT
                                 where  TITLE = new.PROJECT)
         ) $;$
-- 5. close the current state --
   update H_EMPLOYEE
   set    end = new.start 
   where  CODE = old.CODE and end = '$Tfuture$' $;$
-- 6. insert new current state in H_EMPLOYEE (CODE preserved) --
   insert into H_EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) 
   values (old.CODE,new.start,new.NAME,new.STATUS,new.CITY,new.PROJECT) $;$
-- 7. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('update','EMPLOYEE',new.start,old.CODE,
                 new.NAME||', '||new.STATUS||', '||new.CITY||', '||new.PROJECT,'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;

-- delete an employee --
-- ------------------ --
create trigger TRG_EMP_DELETE instead of update on EMPLOYEE
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from H_EMPLOYEE
                      where  CODE = old.CODE
                      and    end = '$Tfuture$') $;$
-- 3. close the current state --
   update H_EMPLOYEE
   set    end = new.end 
   where  CODE = old.CODE and end = '$Tfuture$' $;$
-- 4. Write success in the journal --
   insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG)
          values('delete','EMPLOYEE',new.end,old.CODE,null,'OK') $;$
-- Failure must be detected and coped with in the calling script --
end;

closeDB;


-- ------------------------------------------------------ --
-- LOADING INITIAL STATE OF THE VALID-TIME DATABASE       -- --------------------------------------------------------- --

if ($onerror$ = 1) onError stop;
if ($onerror$ = 2) onError continue;

parameter commitmode = autocommit;

openDB TDB.db;

write;

insert into PROJECT(TITLE,start,THEME,BUDGET) values('BIOTECH','2014-11-18','Biotechnology',180000);
       --write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','PROJECT','2014-11-18','BIOTECH','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('G96','2014-11-23','Godin','F','Genve','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'G96';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2014-11-23','G96','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('M158','2014-12-04','Mercier','F','Paris','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'M158';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2014-12-04','M158','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: duplicate PK --
  insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('G96','2014-12-04','Mercier','F','Paris','BIOTECH');
         write $SQLdiag$ - $EXTENDEDdiag$ 'G96';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2014-12-04','G96','<$EXTENDEDdiag$>','FAILED');
endif;

update EMPLOYEE set start = '2015-01-17',STATUS = 'P' where CODE = 'G96';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'G96';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-01-17','G96','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2015-02-21',STATUS = 'P' where CODE = 'M158';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'M158';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-02-21','M158','<$EXTENDEDdiag$>','FAILED');

update PROJECT set start = '2015-02-27',THEME = 'Genetic engineering', BUDGET = 160000 where TITLE = 'BIOTECH';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2015-02-27','BIOTECH','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('A68','2015-06-09','Albert','P','Mons','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'A68';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2015-06-09','A68','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: no modification --
  update PROJECT set start = '2015-06-15',BUDGET = 160000 where TITLE = 'BIOTECH';
         write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2015-06-15','BIOTECH','<$EXTENDEDdiag$>','FAILED');
endif;

update PROJECT set start = '2015-06-15',BUDGET = 120000 where TITLE = 'BIOTECH';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2015-06-15','BIOTECH','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: FK --
  update PROJECT set end = '2015-06-18' where TITLE = 'BIOTECH';
         write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','PROJECT','2015-06-18','BIOTECH','<$EXTENDEDdiag$>','FAILED');
endif;

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('D122','2015-06-15','Declercq','P','Mons','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D122';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2015-06-15','D122','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('C45','2015-08-22','Carlier','F','Lille','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'C45';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2015-08-22','C45','<$EXTENDEDdiag$>','FAILED');

insert into PROJECT(TITLE,start,THEME,BUDGET) values('SURVEYOR','2015-10-30','Satellite monitoring',310000);
       --write $SQLdiag$ - $EXTENDEDdiag$ 'SURVEYOR';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','PROJECT','2015-10-30','SURVEYOR','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2015-10-30',PROJECT = 'SURVEYOR' where CODE = 'A68';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'A68';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-10-30','A68','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: FK error --
  update EMPLOYEE set start = '2015-10-30',PROJECT = 'SSURVEYORR' where CODE = 'M158';
         write $SQLdiag$ - $EXTENDEDdiag$ 'M158';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-10-30','M158','<$EXTENDEDdiag$>','FAILED');
endif;

update EMPLOYEE set start = '2015-10-30',PROJECT = 'SURVEYOR' where CODE = 'M158';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'M158';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-10-30','M158','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('D107','2016-02-16','Delecourt','F','Grenoble','SURVEYOR');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D107';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2016-02-16','D107','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-03-21',PROJECT = 'SURVEYOR' where CODE = 'D122';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D122';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2015-02-21','D122','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-03-22',CITY = 'Genve' where CODE = 'D107';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D107';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-03-22','D107','<$EXTENDEDdiag$>','FAILED');

update PROJECT set start = '2016-06-25',BUDGET = 375000 where TITLE = 'SURVEYOR';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'SURVEYOR';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2016-06-25','SURVEYOR','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('N240','2016-08-18','Nguyen','F','Toulouse','BIOTECH');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'N240';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2016-08-18','N240','$EXTENDEDdiag$>','FAILED');

update PROJECT set start = '2016-08-23',BUDGET = 140000 where TITLE = 'BIOTECH';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2016-08-23','BIOTECH','<$EXTENDEDdiag$>','FAILED');

insert into PROJECT(TITLE,start,THEME,BUDGET) values('AGRO-2000','2016-08-31','Crop improvement',65000);
       --write $SQLdiag$ - $EXTENDEDdiag$ 'AGRO-2000';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','PROJECT','2016-08-31','AGRO-2000','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-08-31',PROJECT = 'AGRO-2000' where CODE = 'C45';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'C45';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-08-31','C45','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-08-31',PROJECT = 'AGRO-2000' where CODE = 'G96';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'G96';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-08-31','G96','<$EXTENDEDdiag$>','FAILED');

insert into EMPLOYEE(CODE,start,NAME,STATUS,CITY,PROJECT) values ('A237','2016-09-16','Antoine','P','Grenoble','AGRO-2000');
       --write $SQLdiag$ - $EXTENDEDdiag$ 'A237';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','EMPLOYEE','2016-09-16','A237','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-09-16', CITY = 'Grenoble',PROJECT = 'SURVEYOR' where CODE = 'N240';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'N240';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-09-16','N240','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-10-22',CITY = 'Paris' where CODE = 'A68';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'A68';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-10-22','A68','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-12-08',STATUS = 'P',CITY = 'Grenoble' where CODE = 'D107';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D107';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-12-08','D107','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-12-11',PROJECT = 'BIOTECH' where CODE = 'C45';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'C45';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-12-11','C45','<$EXTENDEDdiag$>','FAILED');

update PROJECT set start = '2016-12-20',BUDGET = 345000 where TITLE = 'SURVEYOR';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'SURVEYOR';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','PROJECT','2016-12-20','SURVEYOR','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2016-12-29',CITY = 'Genve',PROJECT = 'AGRO-2000' where CODE = 'N240';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'N240';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2016-12-29','N240','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2017-01-14',PROJECT = 'BIOTECH' where CODE = 'M158';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'M158';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2017-01-14','M158','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2017-04-03',PROJECT = 'BIOTECH' where CODE = 'D107';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D107';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2017-04-03','D107','<$EXTENDEDdiag$>','FAILED');

update PROJECT set start = '2017-04-14',BUDGET = 75000 where TITLE = 'AGRO-2000';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'AGRO-2000';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('insert','PROJECT','2017-04-14','AGRO-2000','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set start = '2017-05-19',CITY = 'Charleroi',PROJECT = 'AGRO-2000' where CODE = 'D122';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D122';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','EMPLOYEE','2017-05-19','D122','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set end = '2017-05-19' where CODE = 'A68';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'A68';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','EMPLOYEE','2017-05-19','A68','<$EXTENDEDdiag$>','FAILED');

update PROJECT  set end = '2017-05-19' where TITLE = 'SURVEYOR';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'SURVEYOR';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','PROJECT','2017-05-19','SURVEYOR','<$EXTENDEDdiag$>','FAILED');

update PROJECT  set start = '2017-07-05',BUDGET = 82000 where TITLE = 'AGRO-2000';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'AGRO-2000';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2017-07-05','AGRO-2000','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: invalid end date --
  update EMPLOYEE set end = '2017-13-24' where CODE = 'D122';
         write $SQLdiag$ - $EXTENDEDdiag$ 'D122';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','EMPLOYEE','2017-13-24','D122','<$EXTENDEDdiag$>','FAILED');
endif;

update EMPLOYEE set end = '2017-07-24' where CODE = 'D122';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'D122';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','EMPLOYEE','2017-07-24','D122','<$EXTENDEDdiag$>','FAILED');

update EMPLOYEE set end = '2017-08-09' where CODE = 'N240';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'N240';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('delete','EMPLOYEE','2017-08-09','N240','<$EXTENDEDdiag$>','FAILED');

if ($errors$ = 1);
  -- ERROR: invalid start date --
  update PROJECT  set start = '2017-09-37',THEME = 'Biotechnologie' where TITLE = 'BIOTECH';
         write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
         if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2017-09-37','BIOTECH','<$EXTENDEDdiag$>','FAILED');
endif;

update PROJECT  set start = '2017-09-17',THEME = 'Biotechnology' where TITLE = 'BIOTECH';
       --write $SQLdiag$ - $EXTENDEDdiag$ 'BIOTECH';
       if ('$SQLdiag$' <> 'OK') insert into LOG(OPER,TYPE,START,ID,MESSAGE,DIAG) values('update','PROJECT','2017-09-17','BIOTECH','<$EXTENDEDdiag$>','FAILED');

commitDB;

onError stop;

write-b -- H_PROJECT --;
select * from H_PROJECT order by TITLE,start;
write-b -- PROJECT --;
select * from PROJECT order by TITLE;

write-b -- H_EMPLOYEE --;
select * from H_EMPLOYEE order by CODE,start;
write-b -- EMPLOYEE --;
select * from EMPLOYEE order by CODE;

write-b -- LOG --;
select * from LOG;

closeDB;

parameter commitmode = default;

write-b End of script;

return;
