-- Database schema for valid time with granularity "date" --
-- Event times are computed as the last date + a random increment
-- between, say, 0 and 10.

-- J-L Hainaut - May 5, 2019


createOrReplaceDB TDB.db;

write-b -- ------------------------------------------------------- --;
write   -- CREATE VALID-TIME EVENT-BASED TEMPORAL DATABASE         --;
write   --  Small version: 3 projects, 8 employees                 --;
write-a -- ------------------------------------------------------- --;


-- Parameters --
-- ---------- --
-- timestamp of future
   set Tfuture = 9999-12-31;


create table EH_PROJECT(
   TITLE    char(20)   not null,
   THEME    char(22)   not null,
   BUDGET   decimal(8) not null,
   Time     integer    not null,
   Event    char(6)    not null,
   primary key (TITLE,Time));

create table EH_EMPLOYEE(
   CODE     char(5)  not null,
   NAME     char(10) not null,
   SALARY   integer  not null,
   CITY     char(10) not null,
   PROJECT  char(20) not null,
   Time     integer  not null,
   Event     char(6) not null,
   primary key (CODE,Time));


-- Current projects --
-- ---------------- --

create view PROJECT(TITLE,THEME,BUDGET,start,end)
as select TITLE,THEME,BUDGET,Time,'$Tfuture$'
   from   EH_PROJECT P
   where  P.Event in ('create','update')
   and    not exists(select * from EH_PROJECT
                     where  TITLE = P.TITLE
                     and    Time > P.Time);

-- Current employees --
-- ----------------- --

create view EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start,end)
as select CODE,NAME,SALARY,CITY,PROJECT,Time,'$Tfuture$'
   from   EH_EMPLOYEE E
   where  E.Event in ('create','update')
   and    not exists(select * from EH_EMPLOYEE
                     where  CODE = E.CODE
                     and    Time > E.Time);


-- Same as EH_PROJECT where time ID have been replaced with true timestamp --
-- ----------------------------------------------------------------------- --


-- Same as EH_EMPLOYEE where time ID have been replaced with true timestamp --
-- ------------------------------------------------------------------------ --

-- Allow ';' at the end of a query in the body of a trigger --
set ; = ;;

-- ------------------ --
-- PROJECT MANAGEMENT --
-- ------------------ --

-- Start a project --
-- --------------- --
create trigger TRG_PRO_INSERT instead of insert on PROJECT
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from EH_PROJECT where TITLE = new.TITLE) $;$
-- 2. insert data in EH_PROJECT with the last time ID --
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event) 
   values (new.TITLE,new.THEME,new.BUDGET,new.start,'create') $;$
end;

-- Update a project --
-- ---------------- --
create trigger TRG_PRO_UPDATE instead of update on PROJECT
when  not (new.THEME = old.THEME and new.BUDGET = old.BUDGET)
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. insert new current state in EH_PROJECT (TITLE preserved) --
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event) values
      (old.TITLE,new.THEME,new.BUDGET,new.start,'update') $;$
end;

-- Stop a project --
-- -------------- --
create trigger TRG_PRO_DELETE instead of update on PROJECT
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from PROJECT
                      where  TITLE = old.TITLE
                      and    end = '$Tfuture$') $;$
-- 2. There can be no dependent EMPLOYEE entities anymore --
   select raise(ABORT,'dependent EMPLOYEEs still exist')
   where  exists(select * from EMPLOYEE
                 where  PROJECT = old.TITLE) $;$
-- 3. insert a delete event
   insert into EH_PROJECT(TITLE,THEME,BUDGET,Time,Event)
      select TITLE,THEME,BUDGET,new.end,'delete'
      from   PROJECT where TITLE = old.TITLE $;$
end;


-- ------------------- --
-- EMPLOYEE MANAGEMENT --
-- ------------------- --

-- Create an employee --
-- ------------------ --
create trigger TRG_EMP_INSERT instead of insert on EMPLOYEE
begin
-- 1. The Entity PK must be new --
   select raise(ABORT,'duplicate entity PK')
   where  exists(select * from EH_EMPLOYEE where CODE = new.CODE) $;$
-- 2. The target entity of an Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from PROJECT
                     where  TITLE = new.PROJECT) $;$
-- 3. insert data in EH_EMPLOYEE with the last time ID
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event) 
   values (new.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,new.start,'create') $;$
end;

-- Update an employee --
-- ------------------ --
create trigger TRG_EMP_UPDATE instead of update on EMPLOYEE
when not (new.NAME = old.NAME and new.SALARY = old.SALARY
          and new.CITY = old.CITY and new.PROJECT = old.PROJECT)
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 3. The target entity of the new Entity FK must exist --
   select raise(ABORT,'target PROJECT does not exist')
   where  not exists(select * from PROJECT
                     where  TITLE = new.PROJECT) $;$
-- 4. insert new current state in EH_EMPLOYEE (CODE preserved) --
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event) values
      (old.CODE,new.NAME,new.SALARY,new.CITY,new.PROJECT,new.start,'update') $;$
end;

-- Delete an employee --
-- ------------------ --
create trigger TRG_EMP_DELETE instead of update on EMPLOYEE
when new.end < '$Tfuture$'
begin
-- 1. The Entity must exist --
   -- Otherwise, no diagnostic --
-- 2. "new.end" is a valid date --
   select raise(ABORT,'invalid end date')
   where  isNotDate(new.end)
   or     new.end <= (select start from EMPLOYEE
                      where  CODE = old.CODE
                      and    end = '$Tfuture$') $;$
-- 3. insert a delete event
   insert into EH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,Time,Event)
      select CODE,NAME,SALARY,CITY,PROJECT,new.end,'delete'
      from   EMPLOYEE where CODE = old.CODE $;$
end;

closeDB;


-- ------------------------------------------------------ --
-- LOADING INITIAL STATE OF THE TRANSACTION-TIME DATABASE --
-- ------------------------------------------------------ --

parameter commitmode = transaction;

openDB TDB.db;

insert into PROJECT(TITLE,THEME,BUDGET,start) values('BIOTECH','Biotechnology',180000,'2014-11-18');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('G96','Godin',3400,'Genve','BIOTECH','2014-11-23');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('M158','Mercier',2900,'Paris','BIOTECH','2014-12-04');
update EMPLOYEE set start = '2015-01-17',SALARY = 3300 where CODE = 'G96';
update EMPLOYEE set start = '2015-02-21',SALARY = 3000 where CODE = 'M158';
update PROJECT set start = '2015-02-27',THEME = 'Genetic engineering', BUDGET = 160000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('A68','Albert',3200,'Toulouse','BIOTECH','2015-06-09');
update PROJECT set start = '2015-06-15',BUDGET = 120000 where TITLE = 'BIOTECH';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('D122','Declercq',3200,'Paris','BIOTECH','2015-06-15');
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('C45','Carlier',3100,'Lille','BIOTECH','2015-08-22');
insert into PROJECT(TITLE,THEME,BUDGET,start) values('SURVEYOR','Satellite monitoring',310000,'2015-10-30');
update EMPLOYEE set start = '2015-10-30',SALARY = 3700,PROJECT = 'SURVEYOR' where CODE = 'A68';
update EMPLOYEE set start = '2015-10-30',SALARY = 3600,PROJECT = 'SURVEYOR' where CODE = 'M158';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('D107','Delecourt',3800,'Grenoble','SURVEYOR','2016-02-16');
update EMPLOYEE set start = '2016-03-21',SALARY = 3800,PROJECT = 'SURVEYOR' where CODE = 'D122';
update EMPLOYEE set start = '2016-03-22',SALARY = 4100,CITY = 'Genve' where CODE = 'D107';
update PROJECT set start = '2016-06-25',BUDGET = 375000 where TITLE = 'SURVEYOR';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('N240','Nguyen',3100,'Toulouse','BIOTECH','2016-08-18');
update PROJECT set start = '2016-08-23',BUDGET = 140000 where TITLE = 'BIOTECH';
insert into PROJECT(TITLE,THEME,BUDGET,start) values('AGRO-2000','Crop improvement',65000,'2016-08-31');
update EMPLOYEE set start = '2016-08-31',SALARY = 3400,PROJECT = 'AGRO-2000' where CODE = 'C45';
update EMPLOYEE set start = '2016-08-31',SALARY = 3900,PROJECT = 'AGRO-2000' where CODE = 'G96';
insert into EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start) values ('A237','Antoine',4000,'Grenoble','AGRO-2000','2016-09-16');
update EMPLOYEE set start = '2016-09-16',SALARY = 3700,CITY = 'Grenoble',PROJECT = 'SURVEYOR' where CODE = 'N240';
update EMPLOYEE set start = '2016-10-22',SALARY = 3900,CITY = 'Paris' where CODE = 'A68';
update EMPLOYEE set start = '2016-12-08',CITY = 'Grenoble' where CODE = 'D107';
update EMPLOYEE set start = '2016-12-11',SALARY = 3100,PROJECT = 'BIOTECH' where CODE = 'C45';
update EMPLOYEE set start = '2016-12-11',SALARY = 4100 where CODE = 'G96';
update EMPLOYEE set start = '2016-12-11',SALARY = 3800,PROJECT = 'BIOTECH' where CODE = 'A237';
update PROJECT set start = '2016-12-20',BUDGET = 345000 where TITLE = 'SURVEYOR';
update EMPLOYEE set start = '2016-12-29',CITY = 'Genve',PROJECT = 'AGRO-2000' where CODE = 'N240';
update EMPLOYEE set start = '2016-12-29',PROJECT = 'AGRO-2000' where CODE = 'A237';
update EMPLOYEE set start = '2017-01-14',SALARY = 3000,PROJECT = 'BIOTECH' where CODE = 'M158';
update EMPLOYEE set start = '2017-04-03',SALARY = 3300,PROJECT = 'BIOTECH' where CODE = 'D107';
update PROJECT set start = '2017-04-14',BUDGET = 75000 where TITLE = 'AGRO-2000';
update EMPLOYEE set start = '2017-05-19',SALARY = 4100,CITY = 'Toulouse',PROJECT = 'AGRO-2000' where CODE = 'D122';
update EMPLOYEE set start = '2017-05-19',SALARY = 3900 where CODE = 'G96';
-- delete EMPLOYEE --
update EMPLOYEE set end = '2017-05-19' where CODE = 'A68';
-- delete PROJECT --
update PROJECT  set end = '2017-05-19' where TITLE = 'SURVEYOR';
update PROJECT set start = '2017-07-05',BUDGET = 82000 where TITLE = 'AGRO-2000';
-- delete EMPLOYEE --
update EMPLOYEE set end = '2017-07-24' where CODE = 'D122';
-- delete EMPLOYEE --
update EMPLOYEE set end = '2017-08-09' where CODE = 'N240';
update PROJECT set start = '2017-09-17',THEME = 'Biotechnology' where TITLE = 'BIOTECH';

commitDB;


-- Rebuild H_PROJECT from EH_PROJECT  --
-- ---------------------------------- --

create view VEH_PROJECT(TITLE,THEME,BUDGET,start,end) as
  -- Generate all past states --
  select P1.TITLE,P1.THEME,P1.BUDGET,P1.Time,P2.Time
  from   EH_PROJECT P1,EH_PROJECT P2
  where  P1.TITLE = P2.TITLE and P1.Time < P2.time
  and    P1.Event in ('create','update')
  and    not exists(select * from EH_PROJECT
                    where  TITLE = P1.TITLE
                    and    Time > P1.Time and Time < P2.Time)
     union
  -- Generate current states --
  select P.TITLE,P.THEME,P.BUDGET,P.Time,'$Tfuture$'
  from   EH_PROJECT P
  where  P.Event in ('create','update')
  and    not exists(select * from EH_PROJECT
                    where  TITLE = P.TITLE
                    and    Time > P.Time)
;

-- Rebuild H_EMPLOYEE from EH_EMPLOYEE --
-- ----------------------------------- --

create view VEH_EMPLOYEE(CODE,NAME,SALARY,CITY,PROJECT,start,end) as
  -- Generate all past states --
  select P1.CODE,P1.NAME,P1.SALARY,P1.CITY,P1.PROJECT,P1.Time,P2.Time
  from   EH_EMPLOYEE P1,EH_EMPLOYEE P2
  where  P1.CODE = P2.CODE and P1.Time < P2.Time
  and    P1.Event in ('create','update')
  and    not exists(select * from EH_EMPLOYEE
                    where  CODE = P1.CODE
                    and    Time > P1.Time and Time < P2.Time)
     union
  -- Generate current states --
  select P.CODE,P.NAME,P.SALARY,P.CITY,P.PROJECT,P.Time,'$Tfuture$'
  from   EH_EMPLOYEE P
  where  P.Event in ('create','update')
  and    not exists(select * from EH_EMPLOYEE
                    where  CODE = P.CODE
                    and    Time > P.Time)
;

write;
write-b -- -------------------------------------------------------- --;
write   -- Event-based history                                      --;
write   -- -------------------------------------------------------- --;

write-b --  EH_PROJECT: event history of projects    --;
select * from EH_PROJECT  order by TITLE,Time;

write-b --  EH_EMPLOYEE: event history of employees   --;
select * from EH_EMPLOYEE order by CODE,Time;

write-b --  PROJECT: current projects             --;
select * from PROJECT order by TITLE;

write-b --  EMPLOYEE: current employees            --;
select * from EMPLOYEE order by CODE;


write;
write-b -- -------------------------------------------------------- --;
write   -- Rebuilding entity-based history from event-based history --;
write   -- -------------------------------------------------------- --;

write-b --  VEH_PROJECT: entity history of projects   --;
select * from VEH_PROJECT  order by TITLE,start;

write-b --  VEH_EMPLOYEE: entity history of employees  --;
select * from VEH_EMPLOYEE order by CODE,start;


write;
write-b -- -------------------------------------------------------- --;
write   -- Deriving event-based history from entity-based history   --;
write   -- -------------------------------------------------------- --;

write-b --  Event history of projects   --;
   select TITLE,THEME,BUDGET,start as Time,'create' as Event
   from   VEH_PROJECT P
   where not exists(select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   end = P.start)
      union
   select TITLE,THEME,BUDGET,end as Time,'delete' as Event
   from   VEH_PROJECT P
   where not exists(select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   start = P.end)
   and   end <> '$Tfuture$'
      union
   select TITLE,THEME,BUDGET,start as Time,'update' as Event
   from   VEH_PROJECT P
   where  exists(   select *
                    from  VEH_PROJECT
                    where TITLE = P.TITLE
                    and   end = P.start)
   order by TITLE,Time;


write-b --  Event history of employees   --;
   select CODE,NAME,SALARY,CITY,PROJECT,start as Time,'create' as Event
   from   VEH_EMPLOYEE E
   where not exists(select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   end = E.start)
      union
   select CODE,NAME,SALARY,CITY,PROJECT,end as Time,'delete' as Event
   from   VEH_EMPLOYEE E
   where not exists(select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   start = E.end)
   and   end <> '$Tfuture$'
      union
   select CODE,NAME,SALARY,CITY,PROJECT,start as Time,'update' as Event
   from   VEH_EMPLOYEE E
   where  exists(   select *
                    from  VEH_EMPLOYEE
                    where CODE = E.CODE
                    and   end = E.start)
   order by CODE,Time;

closeDB;

parameter commitmode = default;

write-ab End of script;

return;
