
-- -------------------------------------------------------------------- --
--   Evolution of number of employes per PROJECT and per (PROJECT,CITY) --
--   Applicable to valid and transaction time databases                 --
-- -------------------------------------------------------------------- --

set limit = 100;

set style = 2;
selectOne-u style = [/bEvolution of the number of employees.
                     @nChoose predicative or procedural interval computing:]
                     Predicative:||Procedural:;
if ('$DIALOGbutton$' = 'Cancel') return;

if ($style$ = 1) goto PREDICATIVE;
if ($style$ = 2) goto PROCEDURAL;
return;


-- -------------------
label PREDICATIVE;
-- -------------------

write-b -- ================================================================ --;
write   -- 1. Evolution of the number of employees per PROJECT              --;
write   --    Predicative technique                                         --;
write-a -- ================================================================ --;

openDB TDB.db;

write-b -- ----------------------------------------------------------- --;
write   -- EVENT: distinct events of each project                      --;
write   -- ----------------------------------------------------------- --;

create temp table EVENT as
   select PROJECT,Instant
   from (select PROJECT,start as Instant from H_EMPLOYEE
          union
         select PROJECT,end as Instant from H_EMPLOYEE);

select thousandSep(count(*),',') as "EVENT" from EVENT;
select * from EVENT order by PROJECT,Instant limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- STABLE_INTERVAL: stable intervals of each project           --;
write   -- ----------------------------------------------------------- --;

create temp table STABLE_INTERVAL as
   select PROJECT,
          Instant as Start,
          (select min(Instant)
           from   EVENT
           where  PROJECT = E.PROJECT
           and    Instant > E.Instant) as End
   from   EVENT E
   where  End is not null;

select thousandSep(count(*),',') as "STABLE_INTERVAL" from STABLE_INTERVAL;
select * from STABLE_INTERVAL order by PROJECT,Start limit $limit$;

create index XSI on STABLE_INTERVAL(PROJECT,start);

write-b -- ----------------------------------------------------------- --;
write   -- STABLE_EMPLOYEE: stable intervals of EMPLOYEE per project   --;
write   -- ----------------------------------------------------------- --;

create temp table STABLE_EMPLOYEE as
   select S.PROJECT,S.start,S.end
   from   STABLE_INTERVAL S, H_EMPLOYEE E
   where  E.PROJECT = S.PROJECT
   and    E.start <= S.start and S.start < E.end;

select thousandSep(count(*),',') as "STABLE_EMPLOYEE" from STABLE_EMPLOYEE;
select * from STABLE_EMPLOYEE order by PROJECT,start limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- E_COUNT: number of employees in each stable interval        --;
write   -- Computed from STABLE_EMPLOYEE                               --;
write   -- ----------------------------------------------------------- --;

create temp table E_COUNT as
   select PROJECT,count(*) as Number,start,end
   from   STABLE_EMPLOYEE
   group by PROJECT,start;
   -- should be: "group by PROJECT,Start,End", but 10% faster;

select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,Start limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- E_COUNT: number of employees in each stable interval        --;
write   -- Computed from STABLE_INTERVAL * H_EMPLOYEE                  --;
write   -- ----------------------------------------------------------- --;

drop table E_COUNT;
create temp table E_COUNT as
   select S.PROJECT,count(*) as Number,S.start as "Start",S.end as "End"
   from   H_EMPLOYEE E, STABLE_INTERVAL S
   where  E.PROJECT = S.PROJECT
   and    E.start <= S.start
   and    S.start < E.end
   group by S.PROJECT,S.start;
   -- should be: "group by S.PROJECT,S.Start,S.End", but 10% faster;

select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,Start limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- Reduce value-equivalent intervals of E_COUNT                --;
write   -- ----------------------------------------------------------- --;

function status = LTemp:reduce {select PROJECT,Number
                                from   E_COUNT};

if ($status$ = 0);
   select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
   select * from E_COUNT order by PROJECT,Start limit $limit$;
else;
   write Error in "LTemp:reduce": $status$;
endif;

closeDB;


write-ab;
write-b -- ============================================================ --;
write   -- 2. Evolution of the number of employees per PROJECT and CITY --;
write   --    Predicative technique                                     --;
write-a -- ============================================================ --;

openDB TDB.db;

write-b -- ----------------------------------------------------------- --;
write   -- EVENT: distinct events of each (project,city)               --;
write   -- ----------------------------------------------------------- --;

create temp table EVENT as
   select PROJECT,CITY,Instant
   from (select PROJECT,CITY,start as Instant from H_EMPLOYEE
          union
         select PROJECT,CITY,end as Instant from H_EMPLOYEE);

select thousandSep(count(*),',') as "EVENT" from EVENT;
select * from EVENT order by PROJECT,CITY,Instant limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- STABLE_INTERVAL: stable intervals of each (project,city)    --;
write   -- ----------------------------------------------------------- --;

create temp table STABLE_INTERVAL as
   select PROJECT,CITY,
          Instant as Start,
          (select min(Instant)
           from   EVENT
           where  (PROJECT,CITY) = (E.PROJECT,E.CITY)
           and    Instant > E.Instant) as End
   from   EVENT E
   where  End is not null;

select thousandSep(count(*),',') as "STABLE_INTERVAL" from STABLE_INTERVAL;
select * from STABLE_INTERVAL order by PROJECT,CITY,Start limit $limit$;

create index XSI on STABLE_INTERVAL(PROJECT,CITY,start);

write-b -- -------------------------------------------------------------------- --;
write   -- STABLE_EMPLOYEE: stable intervals of EMPLOYEE in each (project,city) --;
write   -- -------------------------------------------------------------------- --;

create temp table STABLE_EMPLOYEE as
   select S.PROJECT,S.CITY,S.start,S.end
   from   STABLE_INTERVAL S, H_EMPLOYEE E
   where  (E.PROJECT,E.CITY) = (S.PROJECT,S.CITY)
   and    E.start <= S.start and S.start < E.end;

select thousandSep(count(*),',') as "STABLE_EMPLOYEE" from STABLE_EMPLOYEE;
select * from STABLE_EMPLOYEE order by PROJECT,CITY,start limit $limit$;


write-b -- ---------------------------------------------------------------- --;
write   -- E_COUNT: number of employees in each stable interval of EMPLOYEE --;
write   -- Computed from STABLE_EMPLOYEE                                    --;
write   -- ---------------------------------------------------------------- --;

create temp table E_COUNT as
   select PROJECT,CITY,count(*) as Number,start,end
   from   STABLE_EMPLOYEE
   group by PROJECT,CITY,start;
   -- should be: "group by PROJECT,CITY,Start,End", but 10% faster;

select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,CITY,Start limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- E_COUNT: number of employees in each stable interval        --;
write   -- Computed from STABLE_INTERVAL * H_EMPLOYEE                  --;
write   -- ----------------------------------------------------------- --;

drop table E_COUNT;
create temp table E_COUNT as
   select S.PROJECT,S.CITY,count(*) as Number,S.start as "Start",S.end as "End"
   from   H_EMPLOYEE E, STABLE_INTERVAL S
   where  (E.PROJECT,E.CITY) = (S.PROJECT,S.CITY)
   and    E.start <= S.start
   and    S.start < E.end
   group by S.PROJECT,S.CITY,S.start;
   -- should be: "group by S.PROJECT,S.Start,S.End", but 10% faster;

select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,CITY,Start limit $limit$;


write-b -- ----------------------------------------------------------- --;
write   -- Reduce value-equivalent intervals of E_COUNT                --;
write   -- ----------------------------------------------------------- --;

function status = LTemp:reduce {select PROJECT,CITY,Number
                                from   E_COUNT};

if ($status$ = 0);
   select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
   select * from E_COUNT order by PROJECT,CITY,Start limit $limit$;
else;
   write Error in "LTemp:reduce": $status$;
endif;

closeDB;

-- ---------
goto EXIT;
-- ---------



-- -------------------
label PROCEDURAL;
-- -------------------

write-b;
write-b -- ================================================================ --;
write   -- 1. Evolution of the number of employees per PROJECT              --;
write   --    Procedural technique (LTemp:valuedInterval                    --;
write-a -- ================================================================ --;

openDB TDB.db;

function LTemp:valuedInterval {select PROJECT from H_EMPLOYEE
                               group by PROJECT into STABLE_EMPLOYEE};

select thousandSep(count(*),',') as "STABLE_EMPLOYEE" from STABLE_EMPLOYEE;
select * from STABLE_EMPLOYEE order by PROJECT,Start limit $limit$;

create temp table E_COUNT as
   select PROJECT,count(*) as Number,start,end
   from   STABLE_EMPLOYEE group by PROJECT,start;

function LTemp:reduce {select PROJECT,Number from E_COUNT};

write;
select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,Start limit $limit$;

closeDB;


write-b;
write-b -- ================================================================ --;
write   -- 2. Evolution of the number of employees per PROJECT and CITY     --;
write   --    Procedural technique (LTemp:valuedInterval                    --;
write-a -- ================================================================ --;

openDB TDB.db;

function LTemp:valuedInterval {select PROJECT,CITY from H_EMPLOYEE
                               group by PROJECT,CITY into STABLE_EMPLOYEE};

select thousandSep(count(*),',') as "STABLE_EMPLOYEE" from STABLE_EMPLOYEE;
select * from STABLE_EMPLOYEE order by PROJECT,CITY,Start limit $limit$;

create temp table E_COUNT as
   select PROJECT,CITY,count(*) as Number,start,end
   from   STABLE_EMPLOYEE group by PROJECT,CITY,start;

function LTemp:reduce {select PROJECT,CITY,Number from E_COUNT};

write;
select thousandSep(count(*),',') as "E_COUNT" from E_COUNT;
select * from E_COUNT order by PROJECT,CITY,Start limit $limit$;

closeDB;


-- ---------
label EXIT;
-- ---------

write-b End of script;

return;
