
set Tfuture = 9999-12-31;
set limit = 10;

write-b -- ========================================================== --;
write   -- Temporal queries with temporal result                      --;
write   -- Applicable to valid time databases only                    --;
write   -- ========================================================== --;

-- Select a database --
set db = TDB;
ask-u db = [/bSelect a database] 
            Database:[(TDB,TDB-100,TDB-200,TDB-500,TDB-1000,TDB-2000,TDB-5000,TDB-10000,TDB-20000,TDB-50000,TDB-100000,TDB-200000)];
if ('$DIALOGbutton$' = 'Cancel') return;

set db = $db$.db;

if (not fileExists('$dbDirectory$/$db$'));
   showMessage Database "$db$" not found.
               @nPlease load the experimental temporal databases from SQLfast website;
   return;
endif;

openDB $db$;

-- H_EMPLOYEE must be non-empty and valid-time.
extract N = select count(*) from H_EMPLOYEE;
if ($N$ > 0);
   extract T = select typeof(start) from H_EMPLOYEE limit 1;
   if ('$T$' = 'int');
      showMessage WARNING:
                  @n@S3TDB.db is not a valid time database.
                  @n@S3Operation aborted.;
      closeDB;
      return;
   endif;
else;
   showMessage Database empty.@nOperation aborted.;
   closeDB;
   return;
endif;

extract Nemp = select count(distinct CODE) from H_EMPLOYEE;
extract maxDate = select addToDate(max(end),100) from H_EMPLOYEE where end < '$Tfuture$';

write-b -- Size of the working set --;
select thousandSep($Nemp$,',') as "EMPLOYEE",
       thousandSep(count(*),',') as "H_EMPLOYEE"
from H_EMPLOYEE;


write-b -- -------------------------------------------------------------- --;
write   -- Q1: History of employees between 1990-01-01 and 1990-02-01     --;
write   -- -------------------------------------------------------------- --;
start-timer;
create temp table RES as
   select CODE,
          NAME,SALARY,CITY,PROJECT,
          max(start,'1990-01-01') as "Start",
          min(end,'1990-02-01') as "End"
   from   H_EMPLOYEE
   where  start < '1990-02-01' and '1990-01-01' < end 
   order by NAME,"Start","End";
set Q1 = $timer$;
extract Q1size = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;

write-b -- -------------------------------------------------------------- --;
write   -- Q2: Show the life span of five employees                       --;
write   -- -------------------------------------------------------------- --;
start-timer;
create temp table RES as
   select CODE,NAME,min(start) as Enter, 
          case when max(end) = '$Tfuture$' then null else max(end) end as Leave
   from   H_EMPLOYEE
   where  CODE in ('U10595','P10568','B10496','F10174','B10604')
   group by CODE,NAME;
set Q2 = $timer$;
extract Q2size = select count(*) from RES;
select * from RES;
drop table RES;

write-b -- --------------------------------------------------------------- --;
write   -- Q3: Start date, end date and duration of the employees who left --;
write   -- --------------------------------------------------------------- --;
start-timer;
create temp table RES as
   select CODE,NAME, min(start) as "Start", max(end) as "End",
          durationDays(min(start),max(end)) as Duration
   from   H_EMPLOYEE
   group by CODE,NAME
   having max(end) < '$Tfuture$';
set Q3 = $timer$;
extract Q3size = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;

write-b -- --------------------------------------------------------------- --;
write   -- Q3a: Better through CTE                                         --;
write   -- --------------------------------------------------------------- --;
start-timer;
create temp table RES as
   with MinMax(CODE,NAME,MinStart,MaxEnd)
   as (select CODE,NAME, min(start) as MinStart, max(end) as MaxEnd
       from   H_EMPLOYEE
       group by CODE,NAME
       having MaxEnd < '$Tfuture$')
   select CODE,NAME,MinStart as "Start",MaxEnd as "End",
          durationDays(MinStart,MaxEnd) as Duration
   from   MinMax;
set Q3a = $timer$;
extract Q3asize = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;

write-b -- -------------------------------------------------------------- --;
write   -- Q4: When did the SALARY for each project exceeds 6,000         --;
write   -- -------------------------------------------------------------- --;
start-timer;
create temp table RES as
   select PROJECT,CODE,SALARY,start,end
   from   H_EMPLOYEE
   where  SALARY > 6000
   order by PROJECT,SALARY,CODE;
set Q4 = $timer$;
extract Q4size = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;

write-b -- -------------------------------------------------------------- --;
write   -- Q5: What was (and when) the maximum salary for each project?   --;
write   -- -------------------------------------------------------------- --;
-- The original form, based on a correlated sub-query, takes too much time --
-- because the sub-query is evaluated for each row of H_EMPLOYEE.          --
start-timer;
create temp table RES as
   with MAXSALARY(PROJECT,SALARY)
   as (select PROJECT,max(SALARY)
       from   H_EMPLOYEE
       group by PROJECT)
   select PROJECT,SALARY,start,end
   from   H_EMPLOYEE E
   where  SALARY = (select SALARY from MAXSALARY
                    where  PROJECT = E.PROJECT)
   order by PROJECT;
set Q5 = $timer$;
extract Q5size = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;

write-b -- -------------------------------------------------------------- --;
write   -- Q6: Start date and current age of each active project          --;
write   -- -------------------------------------------------------------- --;
start-timer;
create temp table RES as
   with StartAge(PROJECT,MinStart,MaxEnd)
   as (select PROJECT, min(start),'$maxDate$'
       from   H_EMPLOYEE
       group by PROJECT
       having max(end) = '$Tfuture$')
   select PROJECT,MinStart as "Start",
          durationDays(MinStart,MaxEnd) as Age
   from   StartAge;
set Q6 = $timer$;
extract Q6size = select count(*) from RES;
select * from RES limit $limit$;
drop table RES;


write;
write-ab PERFORMANCE SYNTHESIS;

select thousandSep($Nemp$,',') as "#employees", 'Q1' as Query,  $Q1$ as Time,  thousandSep($Q1size$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q2' as Query,  $Q2$ as Time,  thousandSep($Q2size$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q3' as Query,  $Q3$ as Time,  thousandSep($Q3size$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q3a' as Query, $Q3a$ as Time, thousandSep($Q3asize$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q4' as Query,  $Q4$ as Time,  thousandSep($Q4size$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q5' as Query,  $Q5$ as Time,  thousandSep($Q5size$,',') as Rows
  union
select thousandSep($Nemp$,',') as "#employees", 'Q6' as Query,  $Q6$ as Time,  thousandSep($Q6size$,',') as Rows
order by Query;

closeDB;

write-b End of script;

return;
